/**
  *  \file client/dlg-ammo.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-ammo.h"
#include "client/wcost.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "cpluslib/string.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "game/actions/basetech.h"
#include "game/actions/baseammo.h"
#include "game/team.h"
#include "game/reg.h"
#include "ui/prefix.h"

namespace {
    class WAmmoList : public UIListbox {
        GStarbaseBuildAmmoAction& p;
     public:
        WAmmoList(GStarbaseBuildAmmoAction& p);
        ~WAmmoList();

        void drawPart(GfxCanvas& can, int from, int to);
        bool isAccessible(int what);
    };

    class WAmmoInfo : public UIWidget {
        GCargoContainer&   transfer;
        GCargoType         observe;
        SignalConnection   conn;
     public:
        WAmmoInfo(GCargoContainer& transfer, GCargoType observe);
        ~WAmmoInfo();

        void drawContent(GfxCanvas& can);
        void setCargoType(GCargoType observe);
        void getLayoutInfo(LayoutInfo& li);
    };

    class WAmmoWindow : public UIWindow {
        GMultiTransaction         mt;
        GPlanetTransfer&          transfer;
        GStarbaseBuildAmmoAction& action;
        WAmmoList                 list;
        WAmmoInfo                 info;
        WObservingItemCostDisplay dpy;
        ObjectHolder              h;

        void build(int32 amount);
     public:
        WAmmoWindow(GPlanet& p);

        bool handleEvent(const UIEvent& e, bool second_pass);
        void scrolled(int);
        void enter(int);
        void cancel(int);
    };
};

WAmmoList::WAmmoList(GStarbaseBuildAmmoAction& p)
    : UIListbox(0, NUM_TORPS+1, 18*font_ems[FONT_NORMAL], NUM_TORPS+1),
      p(p)
{ }

WAmmoList::~WAmmoList()
{ }

bool
WAmmoList::isAccessible(int what)
{
    return p.canHave(what+1);
}

void
WAmmoList::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        const int n = getTop() + i + 1;
        UIColor::Color c;
        if (int t = p.getTechUpgradeNeeded(n)) {
            if (t > 6 && !allowRegistered(p.getPlanet().getOwner()))
                c = UIColor::tc_Faded;  // FIXME: should be DARK/GRAY, not DARK/DARK
            else
                c = UIColor::tc_Blue;
        } else {
            c = UIColor::tc_Static;
        }
        if (!p.canHave(n))
            c = UIColor::tc_Faded;
        
        string_t what;
        if (p.getAmount(n))
            what += "\xE2\x80\xA2";
        else
            what += ' ';
        if (n <= NUM_TORPS)
            what += getTorp(n).getName();
        else
            what += _("Fighter");

        drawEntry(can, c, i, what);
    }
}

/******************************* WAmmoInfo *******************************/

WAmmoInfo::WAmmoInfo(GCargoContainer& transfer, GCargoType observe)
    : transfer(transfer), observe(observe)
{
    conn = UIBaseWidget::observe(transfer.sig_changed);
}

WAmmoInfo::~WAmmoInfo()
{ }

void
WAmmoInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    int y      = getExtent().y;
    int32 have = transfer.getCargo(observe);
    int32 max  = transfer.getMaxCargo(observe);

    /* Line 1: "Blafasel Starbase" */
    ctx.setColor(UIColor::tc_Static);
    ctx.useFont(*fonts[FONT_TITLE]);
    outTextF(ctx, getExtent().x, y, getExtent().w, transfer.getName());
    y += font_heights[FONT_TITLE] + 4;

    /* Line 2: "You have %d <zomething>" */
    ctx.setSolidBackground();
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, getExtent().x, y, getExtent().w,
             format(observe == el_Fighters
                    ? _("You have %d fighter%!1{s%}.")
                    : _("You have %d torpedo%!1{es%}."),
                    have));
    y += font_heights[FONT_NORMAL];

    /* Line 3: "You have %d left" */
    string_t expl;
    if (have >= max)
        expl = _("This is the maximum amount.");
    else if (max - have <= 5000)
        /* the average player has much fewer torps on a starbase. let
           them rest in peace not knowing there's a limit, most of the
           time. */
        expl = format(_("You have room for %d more."), max - have);
    outTextF(ctx, getExtent().x, y, getExtent().w, expl);

    can.update(getExtent());
}

void
WAmmoInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size =
        font_sizes[FONT_NORMAL].scaledBy(20, 2).movedBy(0, font_heights[FONT_TITLE] + 5);
}

void
WAmmoInfo::setCargoType(GCargoType observe)
{
    this->observe = observe;
    drawWidget();
}

/****************************** WAmmoWindow ******************************/

WAmmoWindow::WAmmoWindow(GPlanet& p)
    : UIWindow(_("Build Torpedoes/Fighters"), BLUE_WINDOW, COLOR_GRAY),
      mt(),
      transfer(mt.addNewSubtransaction(new GPlanetTransfer(p))),
      action(mt.addNewSubtransaction(new GStarbaseBuildAmmoAction(transfer, transfer))),
      list(action),
      info(transfer, action.getCargoTypeFromIndex(list.getCurrent()+1)),
      dpy(action, p, true)
{
    setLayout(UIVBoxLayout::instance5);
    add(list);
    add(info);
    add(dpy);

    UIGroup& hg = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& ok = h.add(new UIButton(1000,  _("OK"), SDLK_RETURN));
    UIButton& esc = h.add(new UIButton(1001, _("Cancel"), SDLK_ESCAPE));
    hg.add(ok);
    hg.add(esc);
    hg.add(h.add(new UISpacer()));
    add(hg);
    add(h.add(new UIPrefixArg()));

    ok.sig_fired.add(this, &WAmmoWindow::enter);
    esc.sig_fired.add(this, &WAmmoWindow::cancel);
    list.sig_moved.add(this, &WAmmoWindow::scrolled);

    pack();
}

void
WAmmoWindow::build(int32 amount)
{
    int index = list.getCurrent()+1;

    // first, if we're trying to build, and don't have sufficient
    // tech, we have to upgrade.
    if (amount == 0)
        return;
    if (amount > 0) {
        if (int tech = action.getTechUpgradeNeeded(index)) {
            // can we upgrade tech?
            // FIXME: this should probably be converted to use GMultiTransaction
            GPlanetTransfer tech_pay(action.getPlanet());
            GStarbaseTechUpgradeAction tech_action(action.getPlanet());
            GStarbaseBuildTransaction ta(tech_action, tech_pay);
            if (!tech_action.setTech(tl_Torpedo, tech)) {
                messageBox(_("Building this torpedo requires you to upgrade your tech level over Tech 6. "
                             "As an unregistered player, you can't do that."),
                           _("Tech Upgrade"));
                ta.doCancel();
                tech_pay.doCancel();
                return;
            }
            if (ta.getError()) {
                messageBox(format(_("Building this torpedo requires a tech upgrade. "
                                    "You do not have the %d megacredits required for the upgrade."),
                                  ta.getCost(el_Money)),
                           _("Tech Upgrade"));
                ta.doCancel();
                tech_pay.doCancel();
                return;
            }
            if (UIMessageBox(format(_("Building this torpedo requires tech %d. "
                                      "Do you want to upgrade for %d megacredits?"),
                                    tech,
                                    ta.getCost(el_Money)),
                             _("Tech Upgrade")).doYesNoDialog())
            {
                ta.doCommit();
                tech_pay.doCommit();
            } else {
                ta.doCancel();
                tech_pay.doCancel();
            }

            // do not immediately buy stuff after upgrading tech
            // FIXME: widgets should figure out itself that it needs to
            // be redrawn
            dpy.drawWidget(false);
            list.drawWidget(false);
            getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
            return;
        }
    }

    // okay, we're building stuff
    action.setAmount(index, action.getAmount(index) + action.limitAmount(index, amount));
    dpy.drawWidget(false);
    list.drawWidget(false);
}

bool
WAmmoWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case '+':
            if (int i = e.key.arg)
                build(i);
            else
                build(1);
            return true;
         case '-':
            if (int i = e.key.arg)
                build(-i);
            else
                build(-1);
            return true;
         case ss_Ctrl + '+':
            build(100);
            return true;
         case ss_Ctrl + '-':
            build(-100);
            return true;
         case ss_Alt + '+':
            build(10000);
            return true;
         case ss_Alt + '-':
            build(-10000);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
WAmmoWindow::scrolled(int)
{
    int what = list.getCurrent() + 1;

    // FIXME: this is ugly. Since getFighterForPlayer() returns the fighter
    // by value, we must manually build a temporary to be able to point a
    // reference at it.
    const GFighter& f = getFighterForPlayer(transfer.getPlanet().getOwner());
    const GWeapon& w = (what > 0 && what <= NUM_TORPS
                        ? static_cast<const GWeapon&>(getTorp(what).getTorpedo())
                        : f);

    dpy.setPartCost(w.getCost());
    dpy.setPartTech(transfer.getPlanet().getBaseTech(tl_Torpedo), w.getTech());

    info.setCargoType(action.getCargoTypeFromIndex(what));
}

void
WAmmoWindow::enter(int)
{
    if (!mt.getError()) {
        mt.doCommit();
        stop(1);
    }
}

void
WAmmoWindow::cancel(int)
{
    mt.doCancel();
    stop(0);
}

void
doBaseBuildAmmo(GPlanet& p)
{
    WAmmoWindow aw(p);
    ui_root->centerWidget(aw);
    ui_root->add(aw);
    ui_root->run();
    ui_root->close(aw);
}
