/**
  *  \file game/actions/basebuild.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/basebuild.h"
#include "util/nls.h"

/** Compute value of a tech level. This is the price for going from tech
    1 (or 0) to \c tech and allows the computation of arbitrary steps. */
static int
getTechCost(int tech)
{
    /* Going from tech i to i+1 costs 100*i mc, thus going from 0 to i
       costs 100*\sum[0,i] i = 100*(i*(i-1)/2) = 50*i*(i-1). */
    return 50 * tech * (tech-1);
}

namespace {
    /** Counting actor. This actor will just count the costs of all our
        orders, by calling GStarbaseBuildAction::addItemCost() etc. */
    class CountingActor : public GStarbaseBuildExecuter {
        GPlanet& planet;
        GCost acc;
     public:
        CountingActor(GPlanet& p) throw();
        void setTech(TTechLevel area, int value);
        void setHullSlot(int index, int value, int);
        void setEngine(int index, int value, int);
        void setBeam(int index, int value, int);
        void setLauncher(int index, int value, int);
        void accountHull(int number, int count, int);

        const GCost& getCost() const
            { return acc; }
    };
}

inline
CountingActor::CountingActor(GPlanet& p) throw()
    : planet(p),
      acc()
{ }

void
CountingActor::setTech(TTechLevel area, int value)
{
    acc.add(el_Money, getTechCost(value) - getTechCost(planet.getBaseTech(area)));
}

void
CountingActor::setHullSlot(int index, int value, int)
{
    ASSERT(index > 0 && index <= NUM_HULLS_PER_PLAYER);
    int hull_id = getTruehull(planet.getOwner(), index);
    int diff = value - planet.getBaseHullStoreSlot(index);
    if (diff) {
        // do this assertion here so users can call setHullSlot for
        // non-existing slots as long as they don't try to change
        // something.
        ASSERT(hull_id != 0);
        acc += (::getHull(hull_id).getCost() * diff);
    }
}

void
CountingActor::accountHull(int number, int count, int)
{
    acc += (::getHull(number).getCost() * count);
}

void
CountingActor::setEngine(int index, int value, int)
{
    ASSERT(index > 0 && index <= NUM_ENGINES);
    acc += (::getEngine(index).getCost() * (value - planet.getBaseEngineStore(index)));
}

void
CountingActor::setBeam(int index, int value, int)
{
    ASSERT(index > 0 && index <= NUM_BEAMS);
    acc += (::getBeam(index).getCost() * (value - planet.getBaseBeamStore(index)));
}

void
CountingActor::setLauncher(int index, int value, int)
{
    ASSERT(index > 0 && index <= NUM_TORPS);
    acc += (::getTorp(index).getCost() * (value - planet.getBaseLauncherStore(index)));
}

/***************************** ExecutingActor ****************************/

namespace {
    /** Executing actor. This actor will write all changes to the planet we're working on. */
    class ExecutingActor : public GStarbaseBuildExecuter {
        GPlanet& planet;
     public:
        ExecutingActor(GPlanet& p) throw();
        void setTech(TTechLevel area, int value);
        void setHullSlot(int index, int value, int);
        void setEngine(int index, int value, int);
        void setBeam(int index, int value, int);
        void setLauncher(int index, int value, int);
        void accountHull(int index, int value, int);
    };
}

inline
ExecutingActor::ExecutingActor(GPlanet& p) throw()
    : planet(p)
{ }

void
ExecutingActor::setTech(TTechLevel area, int value)
{
    planet.setBaseTech(area, value);
}

void
ExecutingActor::setHullSlot(int index, int value, int)
{
    planet.setBaseHullStoreSlot(index, value);
}

void
ExecutingActor::setEngine(int index, int value, int)
{
    planet.setBaseEngineStore(index, value);
}

void
ExecutingActor::setBeam(int index, int value, int)
{
    planet.setBaseBeamStore(index, value);
}

void
ExecutingActor::setLauncher(int index, int value, int)
{
    planet.setBaseLauncherStore(index, value);
}

void
ExecutingActor::accountHull(int /*number*/, int value, int)
{
    ASSERT(value == 0);
}

/************************** GStarbaseBuildAction *************************/

GCost
GStarbaseBuildAction::getCost()
{
    CountingActor ca(planet);
    perform(ca);
    return ca.getCost();
}


/*********************** GStarbaseBuildTransaction ***********************/

GStarbaseBuildTransaction::GStarbaseBuildTransaction(GStarbaseBuildAction& action, GCargoContainer& container)
    : GCargoCostTransaction(container),
      action(action),
      conn(action.getPlanet().sig_changed.add(this, &GStarbaseBuildTransaction::update)),
      conn2(action.sig_changed.add(this, &GStarbaseBuildTransaction::update))
{
    const GPlanet& p = action.getPlanet();
    if (!p.isPlayed())
        throw GError(GError::eNotPlaying, _("You do not own this planet"));
    if (!p.hasBase())
        throw GError(GError::eNoBase, _("The planet does not have a starbase"));
    update();
}

GStarbaseBuildTransaction::~GStarbaseBuildTransaction()
{
    if (!isFinished())
        doCancel();
}

void
GStarbaseBuildTransaction::update()
{
    if (!isFinished() && action.getPlanet().hasBase()) {
        CountingActor ca(action.getPlanet());
        action.perform(ca);
        setCost(ca.getCost());
        sig_cost_changed.raise();
    }
}

GError
GStarbaseBuildTransaction::getError()
{
    if (GError e = action.getError())
        return e;
    return GCargoCostTransaction::getError();
}

void
GStarbaseBuildTransaction::commit()
{
    ASSERT(!getError());
    ExecutingActor xa(action.getPlanet());
    action.perform(xa);
    GCargoCostTransaction::commit();
}

void
GStarbaseBuildTransaction::cancel()
{
    GCargoCostTransaction::cancel();
}
