/**
  *  \file game/ship-form.cc
  *  \brief Ship Formulas and Prediction
  *
  *  This module contains functions to compute ship properties and
  *  predict their behaviour.
  *
  *  \todo GShipTurnPredictor operates on TShip objects, not GShip. 
  *  This is ugly, as it gives all sorts of problems when requiring
  *  access to ship properties not stored in TShip. However, the
  *  external interface is nice enough, so we don't need to convert
  *  this code until we actually implement the additional
  *  functionality in this module (experience support, per-ship
  *  functions).
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cmath>
#include "game/ship.h"
#include "game/mission.h"
#include "game/player.h"
#include "game/config.h"
#include "game/coord.h"
#include "game/planet.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/univ.h"
#include "game/fcode.h"
#include "game/team.h"
#include "game/reg.h"

const int GShipTurnPredictor::MOVEMENT_TIME_LIMIT;

/** Constructor. Create a predictor for a single ship.
    \param univ   universe containing the ship
    \param id     ship Id */
GShipTurnPredictor::GShipTurnPredictor(const GUniverse& univ, int id)
    : id(id), towee_override(0),
      univ(univ), fuel_used(0), turns_computed(0)
{
    if (univ.getShip(id).hasShipData()) {
        univ.getShip(id).getShipData(this->ship);
        ship.owner = univ.getShip(id).getRealOwner();
        valid = true;
    } else {
        valid = false;
    }
}

/** Constructor. Create a predictor for a ship towing another ship.
    \param univ   universe containing the ship
    \param id     ship Id
    \param towee_override predictor for towed ship */
GShipTurnPredictor::GShipTurnPredictor(const GUniverse& univ, int id, GShipTurnPredictor& towee_override)
    : id(id), towee_override(&towee_override),
      univ(univ), fuel_used(0), turns_computed(0)
{
    if (univ.getShip(id).hasShipData()) {
        univ.getShip(id).getShipData(this->ship);
        ship.owner = univ.getShip(id).getRealOwner();
        valid = true;
    } else {
        valid = false;
    }
}

static void
doRefinery(TShip& s, int16& ore, int16& sup)
{
    int n = getHull(s.hull).getFuel() - s.ore_cargo[el_Neutronium];
    if (n > ore)
        n = ore;
    if (n > sup)
        n = sup;
    s.ore_cargo[el_Neutronium] += n;
    sup -= n;
    ore -= n;
}

static int
getCloakFuel(const TShip& s, int turns, int real_owner)
{
    const int cfb = config.CloakFuelBurn(real_owner);
    int fuel = getHull(s.hull).getMass() * cfb / 100;
    if (fuel < cfb)
        fuel = cfb;
    if (turns)
        fuel *= turns;
    return fuel;
}

static int
getEngineLoad(const GUniverse& univ, TShip& ship, TShip* towee_override, bool tow_corr)
{
    int mass = getShipMass(ship);
    if (ship.mission == GMission::msn_Tow && ship.mission_tow > 0 && ship.mission_tow <= NUM_SHIPS) {
        int towee_mass = 0;
        if (towee_override && towee_override->ship_id == ship.mission_tow)
            towee_mass = getShipMass(*towee_override);
        else if (univ.getShip(ship.mission_tow).isExistant())
            // FIXME: range check
            towee_mass = univ.getShip(ship.mission_tow).getMass();
        if (tow_corr)
            towee_mass = 10*(towee_mass/10);
        mass += towee_mass;
    }
    return mass;
}

static int
computeFuelUsage(const GUniverse& univ, TShip& ship, TShip* towee_override,
                 bool grav_acc, double dist)
{
    if (ship.warp <= 0)
        return 0;

    int way = ship.warp*ship.warp;
    if (grav_acc)
        way *= 2;

    int load = getEngineLoad(univ, ship, towee_override, host.isTHost());
    long ff = getEngine(ship.engine).getFuelFactor(ship.warp);

    if (host.isTHost()) {
        // THost formula
        return (ff * (load / 10) * long(dist)) / (10000L * way);
    } else if (!config.UseAccurateFuelModel()) {
        // PHost, standard formula
        return rdivaddl(load, 10, 0) * ff * long(dist) / (10000L * way);
    } else {
        // PHost, "accurate" formula
        return roundToInt(load * (1 - std::exp(-(ff * dist) / (way*100000L))));
    }
}

static void
normalizePosition(TShip& ship)
{
    GPoint new_pos = GPoint(ship.x, ship.y).getNormalized();
    ship.x = new_pos.x;
    ship.y = new_pos.y;
}

/** Compute one turn. This will compute all changes for the ship. It
    will also update the towed ship predictor, if any. */
void
GShipTurnPredictor::computeTurn()
{
    // is this actually a predictable ship?
    if (!valid)
        return;

    // where are we?
    int pid                = univ.getPlanetAt(GPoint(ship.x, ship.y));
    const GShip& real_ship = univ.getShip(id);

//     {+++ Mine Laying +++}
//     IF (s.Neutronium<>0) AND (ComputeMineDrop(@s, md)) THEN BEGIN
//       Dec(s.Torps, md.torps);
//     END;

    // Special Missions I (Super Refit, Self Repair, Hiss, Rob) would go here

    /* Alchemy */
    if (std::memcmp(ship.fcode, "NAL", 3) != 0) {
        if (real_ship.canDoSpecial(hf_MerlinAlchemy)) {
            // Merlin converts 3 Sup -> 1 Min
            int t = 0, d = 0, m = 0;
            int mins = ship.supplies / 3;
            if (allowRegistered(ship.owner)) {
                if (std::memcmp(ship.fcode, "alt", 3) == 0)
                    t = mins;
                else if (std::memcmp(ship.fcode, "ald", 3) == 0)
                    d = mins;
                else if (std::memcmp(ship.fcode, "alm", 3) == 0)
                    m = mins;
                else if (host.isPHost()) {
                    if (std::memcmp(ship.fcode, "nat", 3) == 0)
                        d = m = mins/2;
                    else if (std::memcmp(ship.fcode, "nad", 3) == 0)
                        t = m = mins/2;
                    else if (std::memcmp(ship.fcode, "nam", 3) == 0)
                        t = d = mins/2;
                    else
                        t = d = m = mins/3;
                } else
                    t = d = m = mins/3;
            } else
                t = d = m = mins/3;
            ship.supplies -= 3*(t+d+m);
            ship.ore_cargo[el_Tritanium]  += t;
            ship.ore_cargo[el_Duranium]   += d;
            ship.ore_cargo[el_Molybdenum] += m;
        } else if (real_ship.canDoSpecial(hf_NeutronicRefinery)) {
            // Neutronic refinery converts 1 Sup + 1 Min -> 1 Neu
            doRefinery(ship, ship.ore_cargo[el_Tritanium],  ship.supplies);
            doRefinery(ship, ship.ore_cargo[el_Duranium],   ship.supplies);
            doRefinery(ship, ship.ore_cargo[el_Molybdenum], ship.supplies);
        } else if (real_ship.canDoSpecial(hf_AriesRefinery)) {
            // Aries converts 1 Min -> 1 Neu
            int16 max = 0x7FFF;  // we assume that no ship has more cargo
            doRefinery(ship, ship.ore_cargo[el_Tritanium],  max);
            doRefinery(ship, ship.ore_cargo[el_Duranium],   max);
            doRefinery(ship, ship.ore_cargo[el_Molybdenum], max);
        }
    }

    /* Starbase Missions I (Fix, [Recycle, Load Torps]) */
    if (pid > 0) {
        const GPlanet& p = univ.getPlanet(pid);
        if (p.isExistant()
            && p.hasBase()
            && p.getBaseShipyardOrder() == sya_Fix
            && p.getBaseShipyardId() == id)
        {
            // we're at a base which is fixing us
            ship.damage = 0;
            ship.crew   = getHull(ship.hull).getCrew();
        }
    }

    /* Supply Repair */
    if (ship.damage > 0 && ship.supplies > 0) {
        int max = ship.supplies / 5;
        if (max > ship.damage)
            max = ship.damage;
        ship.damage -= max;
        ship.supplies -= 5*max;
    }

    /* Cloak Fuel Burn */
    if ((real_ship.canDoSpecial(hf_Cloak) || real_ship.canDoSpecial(hf_AdvancedCloak))
        && ship.mission > 0
        && isMissionCloaking(ship.mission, ship.owner))
    {
        int fuel = real_ship.canDoSpecial(hf_AdvancedCloak) ? 0 : getCloakFuel(ship, 0, real_ship.getRealOwner());
        if (ship.ore_cargo[el_Neutronium] <= fuel || ship.damage >= config.DamageLevelForCloakFail())
            ship.mission = 0;
        else
            ship.ore_cargo[el_Neutronium] -= fuel;
    }

    /* out of fuel cannot do any mission */
    if (ship.ore_cargo[el_Neutronium] == 0)
        ship.mission = 0;

    /* damage speed limit */
    if (ship.damage > 0) {
        int limit = (raceId(ship.owner) == 2
                     ? (host.isPHost()
                        ? 15 - ship.damage/10
                        : 14 - ship.damage/10)
                     : 10 - ship.damage/10);
        if (limit < 0)
            limit = 0;
        if (ship.warp > limit)
            ship.warp = limit;
    }

    /* Actual movement here */
    long dist2 = long(ship.waypoint_dx*ship.waypoint_dx) + long(ship.waypoint_dy*ship.waypoint_dy);
    if (real_ship.canDoSpecial(hf_Hyperdrive)
        && std::memcmp(ship.fcode, "HYP", 3) == 0
        && ship.warp > 0
        && dist2 >= 20*20)
    {
        /* It's hyperjumping */
        ship.ore_cargo[el_Neutronium] -= 50;
        fuel_used                     += 50;

        /* If it's jumping, it can't tow. Advance time in towee's
           world anyway. */
        if (ship.mission == GMission::msn_Tow)
            ship.mission = 0;
        if (towee_override)
            towee_override->computeTurn();

        /* Now move that bugger. */
        if (host.isExactHyperjumpDistance2(dist2)) {
            // exact jump. The obscure comparison is a creative
            // rendering of the fact that 360.00 and 340.00 are *NOT*
            // exact jumps in THost, but are in PHost.
            ship.x += ship.waypoint_dx;
            ship.y += ship.waypoint_dy;
        } else {
            // non-exact jump
            double dist = 350 / std::sqrt(double(dist2));
            int mx = int(dist * std::abs(ship.waypoint_dx) + 0.4999999);
            int my = int(dist * std::abs(ship.waypoint_dy) + 0.4999999);
            if (ship.waypoint_dx < 0)
                mx = -mx;
            if (ship.waypoint_dy < 0)
                my = -my;
            ship.x += mx;
            ship.y += my;
        }
        ship.waypoint_dx = ship.waypoint_dy = 0;
        ship.warp = 0;
        normalizePosition(ship);
        // FIXME: gravity wells?
    } else if (dist2 && ship.warp > 0) {
        /* Normal movement */
        // First, compute new position in mx,my
        double dist = std::sqrt(double(dist2));
        int way = ship.warp * ship.warp;
        if (real_ship.canDoSpecial(hf_Gravitonic))
            way *= 2;

        int mx = ship.waypoint_dx, my = ship.waypoint_dy;
        if (dist > way) {
            int dx, dy;
            if (host.isTHost()) {
                // THost movement formulas, from Donovan's
                if (std::abs(mx) > std::abs(my)) {
                    dx = int(double(way) * std::abs(mx) / dist + 0.5);
                    dy = int(double(dx) * std::abs(my) / std::abs(mx) + 0.5);
                } else {
                    dy = int(double(way) * std::abs(my) / dist + 0.5);
                    dx = int(double(dy) * std::abs(mx) / std::abs(my) + 0.5);
                }
                if (mx < 0)
                    dx = -dx;
                if (my < 0)
                    dy = -dy;
            } else {
                // PHost. From docs and source.
                double head = getHeadingRad(mx, my);
                double fx = std::sin(head) * way;
                double fy = std::cos(head) * way;
                dx = int(fx);
                dy = int(fy);
                if (dx != mx)
                    dx += sgn(fx);
                if (dy != my)
                    dy += sgn(fy);
                if (mx == 0)
                    dx = 0;
                if (my == 0)
                    dy = 0;
            }

            // we now have the dx,dy we want to move
            mx = dx;
            my = dy;
            dist = way;         // FIXME: use distFromDX(dx,dy) instead?
        }

        /* Here, we have the position offset in mx,my. Compute fuel
           usage. We first have to move our towee, if any. */
        if (towee_override) {
            if (ship.mission == GMission::msn_Tow) {
                // we assume the tow succeeds. FIXME: be more clever?
                if (towee_override->ship.mission == GMission::msn_Tow)
                    towee_override->ship.mission = 0;
                towee_override->ship.warp = 0;
            }
            towee_override->computeTurn();
        }

        int fuel = computeFuelUsage(univ, ship,
                                    towee_override ? &towee_override->ship : 0,
                                    real_ship.canDoSpecial(hf_Gravitonic),
                                    dist);
        ship.ore_cargo[el_Neutronium] -= fuel;
        fuel_used += fuel;

        // okay, move it.
        ship.x += mx;
        ship.y += my;
        ship.waypoint_dx -= mx;
        ship.waypoint_dy -= my;
        normalizePosition(ship);

        // warp wells suck. (the ship to the planet, of course.)
        int wp_x = ship.x + ship.waypoint_dx;
        int wp_y = ship.y + ship.waypoint_dy;

        if (ship.warp > 1 && !univ.getPlanetAt(GPoint(ship.x, ship.y))) {
            if (int gpid = univ.getGravityPlanetAt(GPoint(ship.x, ship.y))) {
                // okay, there is a planet. Move the ship.
                GPoint ppos = univ.getPlanet(gpid).getPos();
                ship.x = ppos.x;
                ship.y = ppos.y;
                // Now adjust the waypoint. If the waypoint was inside
                // the warp well of the same planet, assume the end of
                // this movement order. Otherwise, when users set a
                // waypoint at the edge of the warp well, the ship
                // would try to get there for ever. This is consistent
                // with the CCScript `MoveTo' command and with what
                // most people expect.
                if (univ.getPlanetAt(GPoint(wp_x, wp_y), true) == gpid) {
                    ship.waypoint_dx = ship.waypoint_dy = 0;
                } else {
                    GPoint new_wp = GPoint(wp_x, wp_y).getNearestTo(GPoint(ship.x, ship.y));
                    ship.waypoint_dx = new_wp.x - ship.x;
                    ship.waypoint_dy = new_wp.y - ship.y;
                }
            }
        }
    } else {
        /* No sensible movement order for this ship. Advance towee's
           time anyway. */
        if (towee_override)
            towee_override->computeTurn();
    }

    /* turn is over. */
    ++turns_computed;
}

/** Compute this ship's movement, until movement is over or time runs
    out. */
void
GShipTurnPredictor::computeMovement()
{
    if (valid) {
        const int final_turn = turns_computed + MOVEMENT_TIME_LIMIT;
        while ((ship.waypoint_dx || ship.waypoint_dy) && turns_computed < final_turn) {
            computeTurn();
            if (ship.ore_cargo[el_Neutronium] < 0)
                ship.ore_cargo[el_Neutronium] = 0;
        }
    }
}

/** Override this ship's position. */
void
GShipTurnPredictor::setPosition(GPoint pt)
{
    pt = pt.getNormalized();
    ship.x = pt.x;
    ship.y = pt.y;
}

/** Override this ship's waypoint. */
void
GShipTurnPredictor::setWaypoint(GPoint pt)
{
    pt = pt.getNearestTo(GPoint(ship.x, ship.y));
    ship.waypoint_dx = pt.x - ship.x;
    ship.waypoint_dy = pt.y - ship.y;
}

/** Override this ship's speed. */
void
GShipTurnPredictor::setSpeed(int warp)
{
    ship.warp = warp;
}

/** Override this ship's engine. */
void
GShipTurnPredictor::setEngine(int engine)
{
    ship.engine = engine;
}

static int
getMovementTurns(GUniverse& univ, int sid)
{
    GShipTurnPredictor pred(univ, sid);
    pred.computeMovement();
    return pred.getTurnCount();
}

/** Set ship to its optimum warp. */
void
setOptimumWarp(GUniverse& univ, int sid)
{
    // FIXME: I'd like this to take a 'const GShip', but we need univ/sid for
    // GShipTurnPredictor, and we need to be able to modify it as ships aren't copyable.
    GShip& sh = univ.getShip(sid);

    /* sort out trivial cases */
    if (sh.isHyperdriving() || (sh.getWaypointDX() == 0 && sh.getWaypointDY() == 0))
        return;

    /* Find out how long it takes to get there at this speed */
    int this_speed = getEngine(sh.getEngine()).getMaxEfficientWarp();
    sh.setWarp(this_speed);
    int this_time = getMovementTurns(univ, sid);

    /* Do we make it in finite time? If not, let user resolve it. */
    if (this_time >= GShipTurnPredictor::MOVEMENT_TIME_LIMIT)
        return;

    /* Find whether we can reach the target with a slower speed */
    while (this_speed > 1) {
        --this_speed;
        sh.setWarp(this_speed);
        if (getMovementTurns(univ, sid) > this_time) {
            /* It's slower, so undo and stop */
            sh.setWarp(this_speed+1);
            break;
        }
    }
}

/****************************** Mine Laying ******************************/

/** Parse mine laying mission.
    \return true if ship has a mine laying mission */
bool
parseMineLayingMission(const GShip& sh, const GUniverse& univ, GMineMissionData& md)
{
    /* Check whether all required values are known */
    if (!sh.getMission().isKnown()
        || !sh.getTowId().isKnown()
        || !sh.getIntId().isKnown()
        || !sh.getTorpType().isKnown()
        || !sh.getTorpLauncherCount().isKnown()
        || sh.getAmmo() < 0)
    {
        return false;
    }

    /* Check whether ship can lay mines */
    int ttyp  = sh.getTorpType();
    int tlaun = sh.getTorpLauncherCount();
    int torps = sh.getAmmo();
    if (ttyp <= 0 || tlaun <= 0 || torps <= 0) {
        return false;
    }

    /* Check mission */
    bool usemix = false, usemdx = false;     // Use mix/mdx fcode?
    bool makeweb = false;                    // Make a web field?
    int race = 0;                            // Lay as this race?
    int reqid = 0;                           // Require this Id?
    int torplimit = 0;                       // Torpedo limit
    const int owner = sh.getRealOwner();
    const int mission = sh.getMission();
    if (mission == GMission::msn_LayMines) {
        /* Lay mines: honors mix, mdx, but no other parameters */
        usemix = usemdx = true;
    } else if (mission == GMission::msn_Special) {
        /* Lay web (race is checked later): honors, mix, mdx, but no other parameters */
        usemix = usemdx = true;
        makeweb = true;
    } else if (!host.isPHost() || !allowRegistered(owner)) {
        /* Following missions only for registered players in PHost */
        /* FIXME: as for fcodes, we could query mission.cc for presence of these missions */
        return false;
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayMines) {
        /* Lay Minefield: honors no fcodes, registered-only */
        torplimit = sh.getIntId();
        race      = sh.getTowId();
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayWeb) {
        /* Lay Web Minefield: honors no fcodes, registered-only */
        torplimit = sh.getIntId();
        race      = sh.getTowId();
        makeweb   = true;
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayMinesIn) {
        /* Lay Mines In: honors miX */
        usemix    = true;
        torplimit = sh.getIntId();
        reqid     = sh.getTowId();
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayWebIn) {
        /* Lay Web Mines In: honors miX */
        usemix    = true;
        torplimit = sh.getIntId();
        reqid     = sh.getTowId();
        makeweb   = true;
    } else {
        /* Not a mine laying mission */
        return false;
    }

    /* Postprocess fcodes */
    if (usemix || usemdx) {
        /* We accept a friendly code if it
               - is not listed in fcodes.cc (assuming that is blank, missing, unmaintained)
           OR  -     works on the ship
                 AND is not restricted to registered ships, or player is registered */
        const string_t       fc  = sh.getFCode();
        const GFCodeList&    fcl = GFCodeList::getInstance();
        GFCodeList::iterator fci = fcl.getFCodeByName(fc);
        bool validfc = (fci == fcl.end()
                        || (fci->worksOn(sh)
                            && ((fci->getFlags() & GFCode::fc_Registered) == 0
                                || allowRegistered(owner))));
        if (validfc && fc.size() == 3) {
            if (usemix && fc[0] == 'm' && fc[1] == 'i') {
                /* miX */
                race = getPlayerIdFromChar(fc[2]);
            } else if (usemdx && fc[0] == 'm' && fc[1] == 'd') {
                /* mdX */
                if (fc[2] == '0')
                    torplimit = 100;
                else if (fc[2] >= '1' && fc[2] <= '9')
                    torplimit = 10*(fc[2]-'0');
                else if (fc[2] == 'h')
                    torplimit = torps/2;
                else if (fc[2] == 'q')
                    torplimit = torps/4;
            }
        }
    }

    /* Postprocess race and limit */
    if (race <= 0 || race > NUM_OWNERS)
        race = owner;
    if (torplimit > 0 && torplimit < torps)
        torps = torplimit;

    /* Check whether this type is allowed */
    if (makeweb) {
        if (raceMissionId(owner) != 7 || !config.AllowWebMines())
            return false;
    } else {
        if (!config.AllowMinefields())
            return false;
    }

    /* Postprocess race, reqid */
    const GMinefieldContainer& mfc = univ.getMinefields();
    if (reqid != 0) {
        /* We're requiring a particular minefield, so check whether it exists
           and matches our parameters. If it doesn't, we refuse to lay mines,
           in the same way PHost does. */
        GMinefield* mf = mfc.get(reqid);
        if (!mf)
            return false;

        if (host.hasAutomaticMineIdentity())
            race = mf->getOwner();

        if (mf->getOwner() != race
            || mf->isWeb() != makeweb
            || distanceSquared(mf->getPos(), sh.getPos()) > mf->getUnitsForLaying())
        {
            return false;
        }
    } else {
        /* We're not requiring a particular minefield, so find one. */
        long closest = 0;
        if (host.hasMinefieldCenterBug()) {
            for (GMinefieldContainer::const_iterator i = mfc.begin(); i != mfc.end(); ++i) {
                if (i->getOwner() == race && i->isWeb() == makeweb) {
                    long dist = distanceSquared(i->getPos(), sh.getPos());
                    if (reqid == 0 || dist < closest) {
                        /* Minefield matches type and is close. We note its
                           Id only when we're inside; set it to zero if we're
                           outside the field and make a new one */
                        reqid = (dist <= i->getUnitsForLaying() ? i->getId() : 0);
                        closest = dist;
                    }
                }
            }
        } else {
            for (GMinefieldContainer::const_iterator i = mfc.begin(); i != mfc.end(); ++i) {
                if (i->getOwner() == race && i->isWeb() == makeweb) {
                    long dist = distanceSquared(i->getPos(), sh.getPos());
                    if (dist <= i->getUnitsForLaying() && (reqid == 0 || dist < closest)) {
                        /* Minefield matches type and is close, and we're inside */
                        reqid = i->getId();
                        closest = dist;
                    }
                }
            }
        }
    }

    /* Find mine laying rate */
    int rate, srate, mrate;
    if (makeweb) {
        srate = config.UnitsPerWebRate(owner);
        mrate = config.UnitsPerWebRate(race);
    } else {
        srate = config.UnitsPerTorpRate(owner);
        mrate = config.UnitsPerTorpRate(race);
    }
    rate = (srate < mrate ? srate : mrate);

    /* We now know that we're laying mines, so fill in the structure */
    md.mine_id   = reqid;
    md.owner     = race;
    md.type      = makeweb;
    md.num_torps = torps;
    md.num_units = int32(torps) * rate * square(sh.getTorpType()) / 100;
    return true;
}

bool
parseMineScoopMission(const GShip& sh, const GUniverse& /*univ*/, GMineMissionData& md)
{
    /* Check whether all required values are known */
    if (!sh.getMission().isKnown()
        || !sh.getTowId().isKnown()
        || !sh.getIntId().isKnown()
        || !sh.getTorpType().isKnown()
        || !sh.getTorpLauncherCount().isKnown()
        || !sh.getBeamType().isKnown()
        || !sh.getBeamCount().isKnown())
    {
        return false;
    }

    /* Check whether ship can scoop mines */
    if (sh.getTorpType() <= 0
        || sh.getTorpLauncherCount() <= 0
        || (host.isPHost()
            && (sh.getBeamType() <= 0
                || sh.getBeamCount() <= 0)))
    {
        return false;
    }

    /* Do we want to scoop? */
    int mission = sh.getMission();
    if (mission == GMission::msn_MineSweep) {
        /* Check for "msc" fcode */
        const GFCodeList&    fcl = GFCodeList::getInstance();
        GFCodeList::iterator fci;
        if (sh.getFCode() == "msc"
            && ((fci = GFCodeList::getInstance().getFCodeByName(sh.getFCode())) == fcl.end()
                || (fci->worksOn(sh)
                    && ((fci->getFlags() & GFCode::fc_Registered) == 0
                        || allowRegistered(sh.getRealOwner())))))
        {
            /* accept */
            md.mine_id   = 0;
            md.owner     = sh.getRealOwner();
            md.type      = false; // irrelevant for scooping
            md.num_torps = 0;     // means: no limit
            md.num_units = 0;     // irrelevant for scooping
            return true;
        } else {
            /* reject */
            return false;
        }
    } else if (host.isPHost() && mission == config.ExtMissionsStartAt() + GMission::pmsn_ScoopTorps) {
        /* PHost mission */
        md.mine_id   = sh.getTowId();
        md.owner     = sh.getRealOwner();
        md.type      = false;
        md.num_torps = sh.getIntId();
        md.num_units = 0;
        return true;
    } else {
        return false;
    }
}
