/**
  *  \file game/sweep.cc
  *  \brief "Sweep" utility routines
  *
  *  The file lists are based upon CCSWEEP 1.06 shipped with PCC 1.x.
  *  Comments show when a file was added to the list, plus an explanation
  *  of what the file is for if known.
  *
  *  Version History of CCSWEEP:
  *  - CCSWEEP 1.06 (20020128)
  *  - CCSWEEP 1.05 (20010326)
  *  - CCSWEEP 1.04(2) (20001001), names itself 1.04, knows a file configX.dat(?)
  *  - CCSWEEP 1.04 (19990924)
  *  - CCSWEEP 1.03 (19990922)
  *  - CCSWEEP 1.01 (19990531)
  *  - CCSWEEP 1.0 (19980412)
  *  - SWEEP 1.0 (19971220)
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/sweep.h"
#include "game/gamedirs.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/struct.h"

/************************** GGameDirManipulator **************************/

/** Constructor. */
GGameDirManipulator::GGameDirManipulator()
    : opt_verbose(false), num_removed_files(0)
{ }

/** Set verboseness. Verboseness is off by default.
    \param flag New flag. If true, all erased files will be logged. */
void
GGameDirManipulator::setVerbose(bool flag)
{
    opt_verbose = flag;
}

/** Get verboseness. */
bool
GGameDirManipulator::isVerbose() const
{
    return opt_verbose;
}

/** Get number of files removed so far. */
int
GGameDirManipulator::getNumRemovedFiles() const
{
    return num_removed_files;
}

/** Remove a file from the game directory.
    \param name File base name */
void
GGameDirManipulator::removeGameFile(string_t name)
{
    if (game_file_dir->eraseNT(name)) {
        // FIXME: eraseNT swallows ENOENT as well as EIO, EPERM, etc.
        // Those should probably be logged as well.
        if (opt_verbose) {
            console.write(LOG_INFO, format(_("Erased file %s."), name));
        }
        ++num_removed_files;
    }
}

/*************************** GSweepRunProcessor **************************/

/** Constructor. */
GSweepRunProcessor::GSweepRunProcessor()
{ }

GSweepRunProcessor::~GSweepRunProcessor()
{ }

void
GSweepRunProcessor::removeFile(string_t name)
{
    GGameDirManipulator::removeGameFile(name);
}

/***************************** GSweepControl *****************************/

/** Constructor. */
GSweepControl::GSweepControl()
    : opt_erase_db(false),
      have_scanned(false),
      players()
{ }

/** Scan game directory. This looks for genX.dat files to figure out
    what players are there. This is the same criterion GDirectoryOverview
    uses). However, unlike GDirectoryOverview, this does not look into
    the files. */
void
GSweepControl::scan()
{
    remaining_players.clear();
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        Ptr<Stream> s = game_file_dir->openFileNT(format("gen%d.dat", i), Stream::C_READ);
        if (s) {
            remaining_players += i;
        }
    }
    have_scanned = true;
}

/** Perform "sweep" operation: erase files. Enumerates all files to erase on the
    given GSweepProcessor. */
void
GSweepControl::processFiles(GSweepProcessor& processor)
{
    // Scan for players
    if (!have_scanned)
        scan();

    // Delete configured player files
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (players.contains(i)) {
            processPlayerFiles(processor, i);
            remaining_players -= i;
        }
    }

    // Delete standard temporary files left over by various clients
    processor.removeFile("temp.pln");
    processor.removeFile("temp.bmp");
    processor.removeFile("temp.dat");
    processor.removeFile("path.dat");
    processor.removeFile("templock.dat");

    // If no players remain, delete some more files
    if (remaining_players.empty()) {
        processor.removeFile("control.dat");      // CCSweep 1.0
        processor.removeFile("init.tmp");         // CCSweep 1.0
        if (opt_erase_db) {
            processor.removeFile("stat.cc");      // CCSweep 1.0 (PCC)
            processor.removeFile("score.cc");     // PCC2
            processor.removeFile("config.cc");    // CCSweep 1.0 (PCC <1.0.17)
            processor.removeFile("config2.cc");   // PCC >1.0.17
            processor.removeFile("hconfig.hst");  // CCSweep 1.03+
            processor.removeFile("rn.dat");       // CCSweep 1.03+ (Winplan race name config)
            processor.removeFile("pconfig.src");  // CCSweep 1.04+
        }
    }
}

/** Perform "Sweep" operation: update index file. This rewrites the
    init.tmp index file required by planets.exe and Winplan. Unlike
    CCSweep 1.x, we rebuild the file from scratch. */
void
GSweepControl::processIndex()
{
    if (!have_scanned)
        scan();

    // Only write the file if it wouldn't be empty. If it would be empty,
    // processFiles will have deleted it.
    if (!remaining_players.empty()) {
        // Build new index. Since it's simple, we build it directly instead of
        // using the normal marshalling functions.
        char new_index[2*NUM_PLAYERS];
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            new_index[2*(i-1)]   = (remaining_players.contains(i));
            new_index[2*(i-1)+1] = 0;
        }

        // Write it
        Ptr<Stream> s = game_file_dir->openFile("init.tmp", Stream::C_CREATE | Stream::C_WRITE);
        s->writeT(new_index, sizeof(new_index));
    }
}

/** Set erase-database flag. If set, also deletes files usually kept
    over various turns (such as history databases), and thus cannot be
    recovered by Unpack. Default is disabled.
    \param flag new value */
void
GSweepControl::setEraseDatabase(bool flag)
{
    opt_erase_db = flag;
}

/** Set players to erase.
    \param set Player set */
void
GSweepControl::setPlayers(GPlayerSet set)
{
    players = set;
}

/** Get player set. */
GPlayerSet
GSweepControl::getPlayers() const
{
    return players;
}

/** Get set of remaining players. Reflects the current state of the
    game directory. That is, before the "Sweep" operation is
    performed, it possibly includes the players to be erased. */
GPlayerSet
GSweepControl::getRemainingPlayers()
{
    if (!have_scanned)
        scan();
    return remaining_players;
}

/** Process one player's files. */
void
GSweepControl::processPlayerFiles(GSweepProcessor& processor, int player)
{
    static const char* const turn_files[] = {
        "bdata%d.dat",          // CCSweep 1.0
        "bdata%d.dis",          // CCSweep 1.0
        "gen%d.dat",            // CCSweep 1.0
        "mdata%d.dat",          // CCSweep 1.0
        "pdata%d.dat",          // CCSweep 1.0
        "pdata%d.dis",          // CCSweep 1.0
        "ship%d.dat",           // CCSweep 1.0
        "ship%d.dis",           // CCSweep 1.0
        "shipxy%d.dat",         // CCSweep 1.0
        "target%d.dat",         // CCSweep 1.0
        "target%d.ext",         // CCSweep 1.01+ (VPUnpack etc.)
        "vcr%d.dat",            // CCSweep 1.0
        "obj%d.cc",             // CCSweep 1.0+ (PCC <1.0)
        "contrl%d.dat",         // CCSweep 1.0 (Winplan)
        "mess35%d.dat",         // CCSweep 1.0 (Winplan)
        "mess%d.dat",           // CCSweep 1.0
        "mt%d.txt",             // CCSweep 1.0+ (Winplan maketurn log)
        "kore%d.dat",           // CCSweep 1.01+
        "skore%d.dat",          // CCSweep 1.01+
        "cp%d.cc",              // CCSweep 1.04(2)+ (PCC <1.0.17)
        "fat%d.trn",            // CCSweep 1.04(2)+ (Winplan Maketurn temp file)
        "temp%d.trn",           // CCSweep 1.04(2)+ (Winplan Maketurn temp file)
        0
    };
    static const char* const db_files[] = {
        "chart%d.cc",           // CCSweep 1.0 (PCC)
        "mess%d.cc",            // CCSweep 1.0 (PCC <1.1.5)
        "fleet%d.cc",           // CCSweep 1.01+ (PCC)
        "team%d.cc",            // CCSweep 1.03+ (PCC)
        "auto%d.dat",           // CCSweep 1.0 (Winplan)
        "notes%d.dat",          // CCSweep 1.0 (Winplan)
        "pref%d.dat",           // CCSweep 1.0 (Winplan)
        "task%d.dat",           // CCSweep 1.0 (Winplan)
        "config%d.dat",         // CCSweep 1.0 (Winplan?)
        "vm%d.cc",              // CCSweep 1.05+ (PCC)
        "cmd%d.txt",            // CCSweep 1.06+ (PCC, VPA)
        "vpa%d.db",             // CCSweep 1.06+ (VPA)
        "msg%d.ini",            // PCC, VPA
        0
    };

    for (int i = 0; turn_files[i] != 0; ++i)
        processor.removeFile(format(turn_files[i], player));

    if (opt_erase_db)
        for (int i = 0; db_files[i] != 0; ++i)
            processor.removeFile(format(db_files[i], player));
}
