/**
  *  \file int/vmio.cc
  *  \brief Interpreter: Virtual Machine I/O
  *
  *  \todo We can probably do without the global IntVMSaveContext /
  *  IntVMLoadContext instance pointers: we must make sure that
  *  IntSubroutineValue's never end up on the runtime stack (this can
  *  be done by zeroing them out when processing a 'callind'
  *  instruction), and we can restore IntExecutionFrameContext's by
  *  passing the IntVMLoadContext instance to loadContexts (and thus
  *  IntDataSegment::load).
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/vmio.h"
#include "cpluslib/assert.h"
#include "int/contextio.h"
#include "int/exec.h"
#include "int/interror.h"
#include "int/process.h"
#include "io/io.h"
#include "io/nullstream.h"
#include "util/except.h"
#include "util/nls.h"

/** VM File Format: Object Types. */
enum {
    otyp_Bytecode = 1,          ///< Bytecode object (procedure).
    otyp_Process  = 2,          ///< Process object.
    otyp_Frame    = 3           ///< Stack frame object.
};

/** VM File Format: BCO flags. */
enum {
    bcofl_Procedure = 1         ///< Set if this is a procedure, clear if it is a function.
};

/** VM File Format: Frame flags. */
enum {
    frfl_WantResult = 1         ///< Set if caller wants a result.
};

/**************************** IntVMSaveContext ***************************/

namespace {
    class SaveObject {
        Stream& s;
        uint32_t obj_header[4]; // type, size, id, nproperties
        uint32_t header_pos;
        uint32_t property_index;
        uint32_t this_property_start;
        std::vector<uint32_t> properties;

        void writeHeader();
     public:
        SaveObject(Stream& s);
        void start(uint32_t type, uint32_t id, uint32_t nprop);
        void end();
        void startProperty(uint32_t count);
        void endProperty();
    };

    inline void
    storeItem(char* p, uint32_t v)
    {
        storeUint32(p, v);
    }

    inline void
    storeItem(char* p, const IntOpcode& op)
    {
        *p++ = op.arg;
        *p++ = op.arg >> 8;
        *p++ = op.minor;
        *p++ = op.major;
    }

    inline void
    storeItem(char* p, IntBCORef bco)
    {
        storeUint32(p, IntVMSaveContext::getCurrentInstance()->addBCO(bco));
    }

    template<typename T>
    void
    writeArray32(Stream& out, const T& obj)
    {
        enum { CHUNK = 128 };
        char buffer[CHUNK*4];
        uint32_t total = obj.size();
        uint32_t index = 0;
        while (index < total) {
            uint32_t now = total - index;
            if (now > CHUNK)
                now = CHUNK;
            for (uint32_t i = 0; i < now; ++i) {
                storeItem(&buffer[4*i], obj[index + i]);
            }
            out.writeT(buffer, 4*now);
            index += now;
        }
    }
}

SaveObject::SaveObject(Stream& s)
    : s(s)
{ }

void
SaveObject::writeHeader()
{
    std::vector<char> buffer;

    // Header
    buffer.resize(4*4);
    storeUint32(&buffer.front(), obj_header, 4);
    s.writeT(&buffer.front(), buffer.size());

    // Property table
    buffer.resize(4*properties.size());
    storeUint32(&buffer.front(), &properties.front(), properties.size());
    s.writeT(&buffer.front(), buffer.size());
}

void
SaveObject::start(uint32_t type, uint32_t id, uint32_t nprop)
{
    ++nprop;
    obj_header[0] = type;
    obj_header[1] = id;
    obj_header[2] = 0;
    obj_header[3] = nprop;
    header_pos = s.getPos();
    properties.clear();
    properties.resize(2*nprop);
    properties[0] = 0;
    properties[1] = 0;
    property_index = 1;
    writeHeader();
}

void
SaveObject::end()
{
    ASSERT(property_index == obj_header[3]);
    uint32_t end_pos = s.getPos();
    obj_header[2] = end_pos - header_pos - 4*4;
    s.seek(header_pos);
    writeHeader();
    s.seek(end_pos);
}

void
SaveObject::startProperty(uint32_t count)
{
    this_property_start = s.getPos();
    properties[2*property_index] = count;
}

void
SaveObject::endProperty()
{
    properties[2*property_index + 1] = s.getPos() - this_property_start;
    ++property_index;
}

static IntVMSaveContext* save_instance;

/** Constructor. Makes a blank save context. */
IntVMSaveContext::IntVMSaveContext()
    : prev_instance(save_instance)
{
    save_instance = this;
}

/** Destructor. */
IntVMSaveContext::~IntVMSaveContext()
{
    ASSERT(save_instance == this);
    save_instance = prev_instance;
}

/** Save a bytecode object.
    \param out  Stream to write to
    \param p    Bytecode object to save
    \param id   Id to assign this BCO */
void
IntVMSaveContext::saveBCO(Stream& out, IntBCORef p, uint32_t id)
{
    SaveObject so(out);
    so.start(otyp_Bytecode, id, 9);

    // Property 1: header (num_labels, flags, min_args, max_args)
    so.startProperty(0);
    char header[8];
    storeUint16(header+0, p->is_procedure * bcofl_Procedure);
    storeUint16(header+2, p->min_args);
    storeUint16(header+4, p->max_args);
    storeUint16(header+6, p->num_labels);
    out.writeT(header, sizeof(header));
    so.endProperty();

    // Property 2: "data" (literals for pushlit, data segment)
    so.startProperty(p->data.size());
    p->data.save(out, p->data.size());
    so.endProperty();

    // Property 3: "names" (names for e.g. pushvar, name list)
    so.startProperty(p->names.getNumNames());
    p->names.save(out, p->names.getNumNames());
    so.endProperty();

    // Property 4: "code" (count = number of instructions, size = 4x count). 32 bit per instruction.
    so.startProperty(p->code.size());
    writeArray32(out, p->code);
    so.endProperty();

    // Property 5: "subrs" (count = number of BCOs, size = 4x count). 32 bit containing Ids.
    so.startProperty(p->subrs.size());
    writeArray32(out, p->subrs);
    so.endProperty();

    // Property 6: "local_names" (predeclared locals, name list)
    so.startProperty(p->local_names.getNumNames());
    p->local_names.save(out, p->local_names.getNumNames());
    so.endProperty();

    // Property 7: "name" (name hint for loading, string)
    so.startProperty(0);
    if (p->name.size())
        out.writeT(p->name.data(), p->name.size());
    so.endProperty();

    // Property 8: "file name" (debug file name, string)
    so.startProperty(0);
    if (p->file_name.size())
        out.writeT(p->file_name.data(), p->file_name.size());
    so.endProperty();

    // Property 9: "line numbers" (count = number of lines, size = 8x count)
    so.startProperty(p->line_numbers.size()/2);
    writeArray32(out, p->line_numbers);
    so.endProperty();

    so.end();
}

/** Save a stack frame object.
    \param out Stream to save to
    \param fr  Stack frame to save */
void
IntVMSaveContext::saveFrame(Stream& out, const IntExecutionFrame& fr)
{
    SaveObject so(out);

    // We don't actually need the frame_sp here (it will be ignored and
    // reconstructed upon load), but it doesn't hurt.
    so.start(otyp_Frame, fr.frame_sp, 3);

    // Property 1: header
    so.startProperty(0);
    char header[20];
    storeUint32(header+0, addBCO(fr.bco));
    storeUint32(header+4, fr.pc);
    storeUint32(header+8, fr.context_sp);
    storeUint32(header+12, fr.exception_sp);
    storeUint32(header+16, fr.want_result * frfl_WantResult);
    out.writeT(header, sizeof(header));
    so.endProperty();

    // Property 2: local values (data segment)
    so.startProperty(fr.local_values.size());
    fr.local_values.save(out, fr.local_values.size());
    so.endProperty();

    // Property 3: local names (name list)
    so.startProperty(fr.local_names.getNumNames());
    fr.local_names.save(out, fr.local_names.getNumNames());
    so.endProperty();

    so.end();
}

/** Save process.
    \param out Stream to save to
    \param exc Process to save */
void
IntVMSaveContext::saveProcess(Stream& out, IntExecutionContext& exc)
{
    SaveObject so(out);
    so.start(otyp_Process, 0, 6);

    // Property 1: header
    so.startProperty(0);
    char header[1];
    header[0] = exc.getPriority();
    out.writeT(header, sizeof(header));
    so.endProperty();

    // Property 2: name (string)
    so.startProperty(0);
    if (exc.process_name.size())
        out.writeT(exc.process_name.data(), exc.process_name.size());
    so.endProperty();

    // Property 3: frames (object array)
    so.startProperty(exc.frames.size());
    for (ptr_vector<IntExecutionFrame>::const_iterator i = exc.frames.begin(); i != exc.frames.end(); ++i)
        saveFrame(out, **i);
    so.endProperty();

    // Property 4: contexts (data segment)
    so.startProperty(exc.contexts.size());
    saveContexts(out, exc.contexts);
    so.endProperty();

    // Property 5: exceptions (counts = number, size = 16xcount)
    so.startProperty(exc.exceptions.size());
    for (ptr_vector<IntExceptionHandler>::const_iterator i = exc.exceptions.begin(); i != exc.exceptions.end(); ++i) {
        char tmp[16];
        storeUint32(tmp+0,  (*i)->frame_sp);
        storeUint32(tmp+4,  (*i)->context_sp);
        storeUint32(tmp+8,  (*i)->value_sp);
        storeUint32(tmp+12, (*i)->pc);
        out.writeT(tmp, sizeof(tmp));
    }
    so.endProperty();

    // Property 6: value stack (data segment)
    so.startProperty(exc.value_stack.size());
    exc.value_stack.save(out, exc.value_stack.size());
    so.endProperty();

    so.end();
}

/** Add bytecode object to serialization. This assigns the BCO an Id
    number which can be used to refer to that BCO. The Id number is
    returned. Adding an object multiple times only saves it once.
    \param p The bytecode object
    \return Id number */
uint32_t
IntVMSaveContext::addBCO(IntBCORef p)
{
    static const uint32_t nil = uint32_t(-1U);

    /* Check whether we know this one */
    std::map<IntBCORef,uint32_t>::iterator it = bco_to_id.find(p);
    if (it != bco_to_id.end()) {
        /* This object is already there. Do we know its Id? */
        if (it->second == nil) {
            /* Id not yet known. This means we are currently busy saving its preconditions,
               meaning that there is a cycle. We cannot serialize that. */
            throw IntError(_("Internal error: attempt to serialize cyclic structure"));
        } else {
            /* It is known. */
            return it->second;
        }
    } else {
        /* We don't know it, so prepare it: save into a memory stream to cause its
           preconditions to be saved (and thus allocated an Id) */
        bco_to_id[p] = nil;
        NullStream ns;
        saveBCO(ns, p, nil);

        /* Remember it */
        uint32_t this_id = bco_list.size();
        bco_to_id[p] = this_id;
        bco_list.push_back(p);
        return this_id;
    }
}

/** Add process object to serialization. Note that, unlike bytecode, a process
    can be saved many times. Loading will then create multiple copies of it.
    \param proc The process object */
void
IntVMSaveContext::addProcess(IntExecutionContext& proc)
{
    process_list.push_back(&proc);

    /* Prepare it by saving into a temporary memory stream: this
       causes its preconditions (=BCOs) to be saved */
    NullStream ns;
    saveProcess(ns, proc);
}

/** Save all pending objects.
    \param out Stream to save to */
void
IntVMSaveContext::save(Stream& out)
{
    for (uint32_t i = 0; i < bco_list.size(); ++i)
        saveBCO(out, bco_list[i], i);
    for (uint32_t i = 0; i < process_list.size(); ++i)
        saveProcess(out, *process_list[i]);
}

/** Get current save context instance.
    \return current instance; 0 if none */
IntVMSaveContext*
IntVMSaveContext::getCurrentInstance()
{
    return save_instance;
}

/**************************** IntVMLoadContext ***************************/

class IntVMLoadContext::ObjectLoader {
    Stream& s;
    SubStream sub;
    uint32_t obj_type, obj_id, obj_size, obj_nprop;
    uint32_t prop_next;
    uint32_t prop_count;
    uint32_t current_property;
    uint32_t obj_next;
    std::vector<uint32_t> prop;

    void checkObjSize(uint32_t needed);
 public:
    ObjectLoader(Stream& s)
        : s(s), sub(s), obj_next(s.getPos())
        { }
    bool readObject();
    uint32_t readProperty();

    Stream& getPropertyStream()
        { return sub; }
    uint32_t getPropertyCount() const
        { return prop_count; }
    uint32_t getObjectType() const
        { return obj_type; }
    uint32_t getObjectId() const
        { return obj_id; }
};

void
IntVMLoadContext::ObjectLoader::checkObjSize(uint32_t needed)
{
    if (needed > obj_size) {
        throw FileFormatException(s, _("Invalid size"));
    }
    obj_size -= needed;
}

bool
IntVMLoadContext::ObjectLoader::readObject()
{
    /* Read header */
    std::vector<char> header;
    header.resize(4*4);
    s.seek(obj_next);
    int n = s.read(&header.front(), header.size());
    if (n != int(header.size())) {
        if (n == 0)
            return false;
        throw FileTooShortException(s);
    }

    obj_type  = getUint32(&header[0]);
    obj_id    = getUint32(&header[4]);
    obj_size  = getUint32(&header[8]);
    obj_nprop = getUint32(&header[12]);
    obj_next += 4*4 + obj_size;

    /* Validate */
    checkObjSize(obj_nprop * 8);

    /* Read property headers */
    header.resize(obj_nprop * 8);
    s.readT(&header.front(), header.size());
    prop.clear();
    prop.reserve(2*obj_nprop);
    for (uint32_t i = 0; i < 2*obj_nprop; ++i)
        prop.push_back(getUint32(&header[4*i]));

    /* Initialize properties and skip first one */
    prop_next = s.getPos();
    current_property = 0;
    readProperty();
    return true;
}

uint32_t
IntVMLoadContext::ObjectLoader::readProperty()
{
    /* Do we have another property? */
    if (2*current_property >= prop.size())
        return 0;

    /* Check property */
    uint32_t prop_size;
    uint32_t prop_id;
    prop_count = prop[2*current_property];
    prop_size  = prop[2*current_property + 1];
    prop_id    = current_property++;
    checkObjSize(prop_size);

    /* Initialize content */
    sub.setRange(prop_next, prop_size);
    sub.seek(0);
    prop_next += prop_size;

    return prop_id;
}

namespace {
    inline void
    fetchItem(const char* p, std::vector<uint32_t>& out)
    {
        out.push_back(getUint32(p));
    }

    inline void
    fetchItem(const char* p, std::vector<IntBCORef>& out)
    {
        IntBCORef bco = IntVMLoadContext::getCurrentInstance()->getBCO(getUint32(p));
        if (!bco)
            throw std::runtime_error(_("Invalid object reference"));
        out.push_back(bco);
    }

    inline void
    fetchItem(const char* p, std::vector<IntOpcode>& out)
    {
        IntOpcode o;
        o.arg   = getUint16(p);
        o.minor = p[2];
        o.major = p[3];
        out.push_back(o);
    }

    template<typename T>
    void
    loadArray32(Stream& s, uint32_t count, T& out)
    {
        enum { CHUNK = 128 };
        char buffer[CHUNK*4];
        uint32_t index = 0;
        while (index < count) {
            uint32_t now = count - index;
            if (now > CHUNK)
                now = CHUNK;
            s.readT(buffer, 4*now);
            for (uint32_t i = 0; i < now; ++i) {
                fetchItem(&buffer[4*i], out);
            }
            index += now;
        }
    }

    void loadString(Stream& s, string_t& out)
    {
        char buffer[128];
        int n;
        while ((n = s.read(buffer, sizeof(buffer))) > 0)
            out.append(buffer, n);
    }
}

static IntVMLoadContext* load_instance;

/** Constructor. Makes a blank instance. */
IntVMLoadContext::IntVMLoadContext()
    : prev_instance(load_instance)
{
    load_instance = this;
}

/** Destructor. */
IntVMLoadContext::~IntVMLoadContext()
{
    ASSERT(load_instance == this);
    load_instance = prev_instance;
}

/** Load bytecode object.
    \param ldr Object loader that has just read the object header
    \return newly-created BCO
    \todo If we find that this BCO is identical to one that is currently
    in the symbol table (using the Name property), recycle that. */
IntBCORef
IntVMLoadContext::loadBCO(ObjectLoader& ldr)
{
    IntBCORef obj = new IntBytecodeObject();
    while (uint32_t id = ldr.readProperty()) {
        switch (id) {
         case 1:
            // Header
            {
                char header[8];
                int n = ldr.getPropertyStream().read(header, sizeof(header));
                if (n >= 2) {
                    uint16_t flags = getUint16(header+0);
                    obj->is_procedure = (flags & bcofl_Procedure) != 0;
                }
                if (n >= 4) {
                    obj->min_args = getUint16(header+2);
                }
                if (n >= 6) {
                    obj->max_args = getUint16(header+4);
                }
                if (n >= 8) {
                    obj->num_labels = getUint16(header+6);
                }
            }
            break;

         case 2:
            // "data" (literals for pushlit, data segment)
            obj->data.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 3:
            // "names" (names for e.g. pushvar, name list)
            obj->names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 4:
            // "code" (count = number of instructions, size = 4x count). 32 bit per instruction.
            loadArray32(ldr.getPropertyStream(), ldr.getPropertyCount(), obj->code);
            break;

         case 5:
            // "subrs" (count = number of BCOs, size = 4x count). 32 bit containing Ids.
            loadArray32(ldr.getPropertyStream(), ldr.getPropertyCount(), obj->subrs);
            break;

         case 6:
            // "local_names" (predeclared locals, name list)
            obj->local_names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 7:
            // "name" (name hint for loading, string)
            loadString(ldr.getPropertyStream(), obj->name);
            break;

         case 8:
            // "file name" (debug file name, string)
            loadString(ldr.getPropertyStream(), obj->file_name);
            break;

         case 9:
            // "line numbers" (count = number of lines, size = 8x count)
            loadArray32(ldr.getPropertyStream(), ldr.getPropertyCount() & ~1, obj->line_numbers);
            break;

         default:
            break;
        }
    }
    return obj;
}

/** Load process. The process will be created in runnable state.
    \param ldr ObjectLoader that has just read the object header */
void
IntVMLoadContext::loadProcess(ObjectLoader& ldr)
{
    IntExecutionContext& proc = createProcess(string_t());

    IntExecutionContext* old_process = current_process;
    current_process = &proc;
    try {
        while (uint32_t id = ldr.readProperty()) {
            switch (id) {
             case 1:
                // header
                {
                    char hdr[1];
                    int n = ldr.getPropertyStream().read(hdr, sizeof(hdr));
                    if (n >= 1) {
                        // FIXME: when we implement process priorities, this must move the process
                        // around in the process list.
                        proc.process_priority = uint8_t(hdr[0]);
                    }
                }
                break;
             case 2:
                // name (string)
                loadString(ldr.getPropertyStream(), proc.process_name);
                break;

             case 3:
                // frames (object array)
                loadFrames(proc, ldr.getPropertyStream(), ldr.getPropertyCount());
                break;

             case 4:
                // contexts (data segment)
                loadContexts(ldr.getPropertyStream(), ldr.getPropertyCount(), proc.contexts);
                break;

             case 5:
                // exceptions (counts = number, size = 16xcount)
                for (uint32_t i = 0, end = ldr.getPropertyCount(); i < end; ++i) {
                    char frame[4*4];
                    ldr.getPropertyStream().readT(frame, sizeof(frame));

                    IntExceptionHandler* ex = new IntExceptionHandler();
                    ex->frame_sp   = getUint32(frame+0);
                    ex->context_sp = getUint32(frame+4);
                    ex->value_sp   = getUint32(frame+8);
                    ex->pc         = getUint32(frame+16);
                    proc.exceptions.push_back(ex);
                }
                break;

             case 6:
                // value stack (data segment)
                proc.value_stack.load(ldr.getPropertyStream(), ldr.getPropertyCount());
                break;
            }
        }
    }
    catch (...) {
        /* If loading fails with an exception, make sure that the process will not run */
        current_process = old_process;
        proc.process_state = proc.psTerminated;
        throw;
    }
    current_process = old_process;
}

/** Load stack frames.
    \param proc  Process to load stack frames into
    \param s     Stream to read from
    \param count Number of stack frames to read */
void
IntVMLoadContext::loadFrames(IntExecutionContext& proc, Stream& s, uint32_t count)
{
    ObjectLoader ldr(s);
    proc.frames.reserve(count);
    while (count-- > 0) {
        // Read frame object
        if (!ldr.readObject())
            throw FileFormatException(s, _("Invalid frame"));
        if (ldr.getObjectType() != otyp_Frame)
            throw FileFormatException(s, _("Invalid frame type"));

        // Read frame content
        bool have_header = false;
        while (uint32_t id = ldr.readProperty()) {
            switch (id) {
             case 1:
                // header
                if (have_header) {
                    throw FileFormatException(s, _("Invalid frame"));
                } else {
                    char tmp[5*4];
                    int n = ldr.getPropertyStream().read(tmp, sizeof(tm));
                    // bco_id is mandatory
                    if (n < 4)
                        throw FileFormatException(s, _("Invalid frame"));
                    proc.frames.push_back(new IntExecutionFrame(getBCO(getUint32(tmp+0))));
                    have_header = true;

                    proc.frames.back()->pc           = getUint32(tmp+4);
                    proc.frames.back()->context_sp   = getUint32(tmp+8);
                    proc.frames.back()->exception_sp = getUint32(tmp+12);

                    uint32_t flags = getUint32(tmp+16);
                    if (flags & frfl_WantResult)
                        proc.frames.back()->want_result = true;
                }
                break;

             case 2:
                // local values (data segment)
                if (!have_header)
                    throw FileFormatException(s, _("Invalid frame"));
                proc.frames.back()->local_values.load(ldr.getPropertyStream(), ldr.getPropertyCount());
                break;

             case 3:
                // local names (name list)
                if (!have_header)
                    throw FileFormatException(s, _("Invalid frame"));
                proc.frames.back()->local_names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
                break;
            }
        }
    }
}

/** Add bytecode object and Id
    \param p  Bytecode object
    \param id Assigned Id */
void
IntVMLoadContext::addBCO(IntBCORef p, uint32_t id)
{
    id_to_bco[id] = p;
}

/** Get bytecode object by Id
    \param id Id to check
    \return bytecode object; 0 if none */
IntBCORef
IntVMLoadContext::getBCO(uint32_t id) const
{
    std::map<uint32_t,IntBCORef>::const_iterator p = id_to_bco.find(id);
    if (p != id_to_bco.end())
        return p->second;
    else
        return 0;
}

/** Make stack frame context by Id. This is used to deserialize stack
    frame contexts. It only works while actually deserializing a process.
    \param id Id read from file
    \return IntExecutionFrameContext object; 0 if those are not allowed here */
IntContext*
IntVMLoadContext::makeFrameContext(uint32_t id) const
{
    if (current_process == 0 || id >= current_process->frames.size())
        return 0;
    return new IntExecutionFrameContext(*current_process->frames[id]);
}

/** Load VM file.
    \param s Stream to read from
    \param acceptProcesses true to accept saved processes (=VM file),
    false to accept only bytecode (=object file) */
void
IntVMLoadContext::load(Stream& s, bool acceptProcesses)
{
    ObjectLoader ldr(s);
    while (ldr.readObject()) {
        if (ldr.getObjectType() == otyp_Process && acceptProcesses) {
            loadProcess(ldr);
        } else if (ldr.getObjectType() == otyp_Bytecode) {
            addBCO(loadBCO(ldr), ldr.getObjectId());
        } else {
            printf("found object of type %d at %d\n", ldr.getObjectType(), s.getPos());
            throw FileFormatException(s, _("Unexpected object"));
        }
    }
}

/** Get current load context instance.
    \return current instance; 0 if none */
IntVMLoadContext*
IntVMLoadContext::getCurrentInstance()
{
    return load_instance;
}
