#!/usr/bin/perl -w
#
#  Check initialisations
#
#  "perl checkinits.pl binary"
#
#  All initializers that appear in the binary must be referenced in
#  its initialisation table. There must be only one such table. This
#  uses the fact that whenever a module is referenced, its module
#  initializer also appears in the binary, even if it's unreferenced.
#  However, even if the linker in one platform would eliminate such
#  unused data, we could still use checkinits.pl on another one where
#  it doesn't, or temporarily disable that option for the check.
#
#  Definitions of module initializers can be preceded by a comment.
#     // Constraints: after mod_foo, before mod_bar, with mod_quux or mod_quuxuum, not with mod_narf
#  Those define the required order of initializers. Keywords are
#  'before', 'after', 'with', 'not' and 'or'.
#
#  This also attempts to check null termination of initializers and
#  tables.
#
#  This uses the shell command 'nm'. It parses the Makefile to find
#  the source file names.
#
#  It assumes that all module initializers are named 'mod_foo', and
#  all initialisation tables are named 'foo_modules'.
#
use strict;

# Options
my $opt_table = 0;
while (@ARGV) {
    if ($ARGV[0] eq '--table') { $opt_table = 1; shift @ARGV }
    else { last }
}

if (@ARGV != 1) {
    die "Usage: $0 [--table] binary\n";
}

# Read symbols that appear in binary
my @mods;
my @modlists;
open NM, "nm $ARGV[0] |" or die;
while (<NM>) {
    if (/(mod_\S+)/) {
        push @mods, $1;
    } elsif (/\b_?(\S+_modules)\b/) {
        push @modlists, $1;
    }
}
close NM;

# Verify module lists
if (@modlists != 1) {
    print "Error: There are ", scalar(@modlists), " module lists in this binary. I expected to find one.\n";
    if (@modlists) {
        print "I found those: ", join(', ', @modlists), "\n";
    }
    if (@mods) {
        print "I expect those modules: ", join(', ', @mods), "\n";
    }
    exit 1;
}

# Read list of source files
my @files;
open MF, "Makefile" or die "Makefile: $!";
while (<MF>) {
    chomp;
    while (s/\\$// && (my $next = <MF>)) {
        $_ .= $next;
        chomp;
    }
    if (/^OBJECTS_\S+\s*=\s*(.*)/) {
        foreach (split /\s+/, $1) {
            if (/(.*)\.o$/) { push @files, "$1.cc" }
        }
    }
}
close MF;
print "(scanning ", scalar(@files), " files...)\n";

# Scan source code to find definitions
my $nameregex = '\b(' . join('|', @mods, @modlists) . ')\b';
my %definitions;            # name -> {file=>file, content=>content, comment=>comment}
foreach my $f (@files) {
    open FILE, "< $f" or die "$f: $!\n";
    my $comment = "";
    my $inblockcomment = 0;
    while (<FILE>) {
        # Process ending block comments
        if ($inblockcomment) {
            if (s|^.*\*/||) { $inblockcomment = 0 }
            else { next }
        }
        if (m|^//+\s*(.*)|) {
            # Gather comment
            $comment .= "$_\n";
        } elsif (m:^(static\s+|extern\s+|struct\s+|const\s+)+ModuleInitializer(\*|const\s+|\s+)*($nameregex)\[\]\s*=\s*\{:o) {
            # Definition
            my $item = $3;
            if (exists($definitions{$item})) {
                print STDERR "$f: duplicate definition for '$item', already in $definitions{$item}{file}\n";
            } else {
                # Read content
                my $content = "";
                while (<FILE>) {
                    if (m|^\s*\#|) { next }
                    if (/^\};/) { last }
                    $content .= $_;
                }
                # Trim content
                $content =~ s|/\*.*?\*/||gs;
                $content =~ s|//.*||gm;
                $content =~ s|[\s\n]+| |gs;
                $content =~ s|^\s+||gs;
                $content =~ s|\s+$||gs;
                $definitions{$item} = { file => $f, content => $content, comment => $comment };
            }
            $comment = "";
        } elsif (m:^(static\s+|extern\s+|struct\s+|const\s+)+ModuleInitializer(\*|const\s+|\s+)*($nameregex)\[\];:o) {
            # Declaration, ignore
        } elsif (m:^(static\s+|extern\s+|struct\s+|const\s+)+ModuleInitializer(\*|const\s+|\s+)*\S+\s*=\s*\{:o) {
            # Skip other definition
            while (<FILE>) {
                if (/^\};/) { last }
            }
            $comment = "";
        } else {
            # Anything else. Warn if it contains one of the keywords
            # we're looking for outside a comment or a
            # performInitialisation invocation.
            s|//.*||;
            if (m/($nameregex)/o && !m/performInitialisation/) {
                print STDERR "$f: warning: uninterpreted reference to '$1':\n";
                print STDERR $_;
            }
            $comment = "";
        }

        # Process starting block comments
        s|/\*.*?\*/||g;
        if (m|/\*|) { $inblockcomment = 1 }
    }
    close FILE;
}

# Validate existence, and build cross-reference table
my $error = 0;
my %xref;
foreach (@modlists) {
    if (!exists($definitions{$_})) {
        die "Error: I didn't find a definition for '$_'.\n";
    }
}
foreach (@mods) {
    if (!exists($definitions{$_})) {
        print STDERR "Warning: I didn't find a definition for '$_'.\n";
    } else {
        if ($definitions{$_}{content} !~ m|\{\s*0\s*,\s*0\s*,?\s*\}\s*,?\s*$|s) {
            # Last entry is not '{ 0, 0 }'
            print STDERR "Error: '$_' is not null-terminated.\n";
            ++$error;
        } elsif ($definitions{$_}{content} =~ m|\{\s*0\s*,\s*0\s*,?\s*\}\s*,\s*\{|s) {
            # There is an entry '{ 0, 0 }, {', i.e. a null entry followed by something
            print STDERR "Error: '$_' contains null entry.\n";
            ++$error;
        }

        my @list = split /,/, $definitions{$_}{content};
        if (@list % 2) {
            print STDERR "Warning: definition for '$_' has an odd number of elements.\n";
        } else {
            foreach (@list) {
                s/[ {},\n\t]+//g;
            }
            pop @list;
            pop @list;
            while (@list) {
                my @items = split /\|/, shift(@list);
                my $function = shift(@list);
                foreach my $i (@items) {
                    push @{$xref{$i}}, "$_/$function";
                }
            }
        }
    }
}

# Check that module list contains all modules
my @usedmodules = split /, */, $definitions{$modlists[0]}{content};
foreach my $m (@mods) {
    if (!grep {$m eq $_} @usedmodules) {
        print STDERR "Error: module '$m' is not listed in '$modlists[0]'\n";
        ++$error;
    }
}

# Check null termination of module list
if ($usedmodules[-1] ne '0') {
    print STDERR "Error: module list '$modlists[0]' is not null-terminated\n";
    ++$error;
} else {
    pop @usedmodules;
}

# Warn for items in the list we don't know. This finds naming
# convention violations and embedded nulls.
foreach my $m (@usedmodules) {
    if (!grep {$m eq $_} @mods) {
        print STDERR "Warning: module list '$modlists[0]' contains item '$m' which I didn't know. Analysis may be unreliable.\n";
    }
}

# Check constraints
foreach my $midx (0 .. $#usedmodules) {
    my $m = $usedmodules[$midx];
    if (exists($definitions{$m})) {
        my $c = $definitions{$m}{comment};
        if ($c =~ /Constraints:\s+(.*)/m) {
            foreach my $e (split /,\s*/, $1) {
                my $negate = 0;
                if ($e =~ s/^not\s*//) { $negate = 1 }
                if ($e =~ /^after\s*(.*)/) {
                    $error += checkConstraint($m, 'after', $negate, $1, @usedmodules[0 .. $midx]);
                } elsif ($e =~ /^before\s*(.*)/) {
                    $error += checkConstraint($m, 'before', $negate, $1, @usedmodules[$midx .. $#usedmodules]);
                } elsif ($e =~ /^with\s*(.*)/) {
                    $error += checkConstraint($m, 'with', $negate, $1, @usedmodules);
                }
            }
        }
    }
}
if ($error) {
    print "$error problems found.\n";
    exit 1;
}
print "No problems found.\n";

if ($opt_table) {
    foreach my $k (sort keys %xref) {
        print "$k:\n";
        foreach (@{$xref{$k}}) {
            print "\t$_\n";
        }
    }
}


# Check a constraint.
#   $m          - current module (for formatting the error message)
#   $kind       - kind of constraint (for formatting the error message)
#   $negate     - true to negate test
#   $check      - operand, i.e. modules to check for
#   @where      - module list to examine
sub checkConstraint {
    my ($m, $kind, $negate, $check, @where) = @_;
    my @checkitems = split /\s+or\s+/, $check;

    # We cannot check function calls
    @checkitems = grep !/\(\)/, @checkitems;
    if (@checkitems == 0) {
        return 0;
    }

    # Build regexp
    my $checkre = join('|', @checkitems);
    $checkre =~ s/\s+or\s+/|/g;
    if (grep /^$checkre$/, @where) {
        # found
        if (!$negate) { return 0 }
    } else {
        # not found
        if ($negate) { return 0 }
    }

    # Error
    my $not = $negate ? 'not ' : '';
    print "Error: constraint '$not$kind $check' of module '$m' not satisfied.\n";
    return 1;
}
