/**
  *  \file u/t_util_cmdline.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/cmdline.h"
#include "u/t_util.h"

namespace {
    class TestCommandLineParser : public CommandLineParser {
        string_t result;
     public:
        TestCommandLineParser(bool traditional)
            : CommandLineParser(traditional)
            { }
        void parseStarted()
            { result += "start\n"; }
        void parseFinished()
            { result += "finish\n"; }
        void handleNonoption(const string_t& no)
            {
                result += "nonoption:";
                result += no;
                result += "\n";
            }
        void handleOption(const string_t& o, bool have_arg)
            {
                result += "option:";
                result += o;
                result += ":";
                result += "ny"[have_arg];
                result += "\n";
            }

        string_t getResult() const
            {
                // printf("result is <<%s>>\n", result.c_str());
                return result;
            }
    };
}

void
UtilCmdlineTestSuite::testCommandLineParser()
{
    /* Very basic case */
    {
        static const char*const a[] = { "prog", 0 };
        TestCommandLineParser tclp(false);
        tclp.parse(a);
        TS_ASSERT_EQUALS(TestCommandLineParser::getProgramName(), "prog");
        TS_ASSERT_EQUALS(tclp.getResult(),
                         "start\n"
                         "finish\n");
    }

    /* ASCII parsing */
    {
        static const char*const a[] = { "prog2", "-foo", "-bar", 0 };
        TestCommandLineParser tclp(false /* not traditional */);
        tclp.parse(a);
        TS_ASSERT_EQUALS(TestCommandLineParser::getProgramName(), "prog2");
        TS_ASSERT_EQUALS(tclp.getResult(),
                         "start\n"
                         "option:foo:n\n"
                         "option:bar:n\n"
                         "finish\n");
    }

    {
        static const char*const a[] = { "prog3", "-foo", "-bar", 0 };
        TestCommandLineParser tclp(true /* traditional */);
        tclp.parse(a);
        TS_ASSERT_EQUALS(TestCommandLineParser::getProgramName(), "prog3");
        TS_ASSERT_EQUALS(tclp.getResult(),
                         "start\n"
                         "option:f:n\n"
                         "option:o:n\n"
                         "option:o:n\n"
                         "option:b:n\n"
                         "option:a:n\n"
                         "option:r:n\n"
                         "finish\n");
    }

    /* UTF-8 parsing */
    {
        static const char*const a[] = { "prog4", "-f\xc2\x85o", "-bar", 0 };
        TestCommandLineParser tclp(false /* not traditional */);
        tclp.parse(a);
        TS_ASSERT_EQUALS(TestCommandLineParser::getProgramName(), "prog4");
        TS_ASSERT_EQUALS(tclp.getResult(),
                         "start\n"
                         "option:f\xc2\x85o:n\n"
                         "option:bar:n\n"
                         "finish\n");
    }
    
    {
        static const char*const a[] = { "prog5", "-f\xc2\x85o", "-bar", 0 };
        TestCommandLineParser tclp(true /* traditional */);
        tclp.parse(a);
        TS_ASSERT_EQUALS(TestCommandLineParser::getProgramName(), "prog5");
        TS_ASSERT_EQUALS(tclp.getResult(),
                         "start\n"
                         "option:f:n\n"
                         "option:\xc2\x85:n\n"
                         "option:o:n\n"
                         "option:b:n\n"
                         "option:a:n\n"
                         "option:r:n\n"
                         "finish\n");
    }

    /* Now use some more features */
    {
        static const char*const a[] = { "prog6", "-foo=2", "-", "--bar", "narf", "--", "-zot", "zonk", 0 };
        TestCommandLineParser tclp(false /* not traditional */);
        tclp.parse(a);
        TS_ASSERT_EQUALS(TestCommandLineParser::getProgramName(), "prog6");
        TS_ASSERT_EQUALS(tclp.getResult(),
                         "start\n"
                         "option:foo:y\n"
                         "nonoption:-\n"
                         "option:bar:n\n"
                         "nonoption:narf\n"
                         "nonoption:-zot\n"
                         "nonoption:zonk\n"
                         "finish\n");
    }
}
