/**
  *  \file ui/widgets/alignwidget.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/alignwidget.h"

/** Compute layout.
    \param cs       [out] Room for two integers; index 0 will receive the (relative) start coordinate,
                    index 1 will receive the client size
    \param avail    [in] Number of pixels we have
    \param minsize  [in] Minimum size requested by client
    \param prefsize [in] Preferred size requested by client
    \param align    [in] Alignment parameter, 0=left, 1=center, 2=right
    \param margin   [in] Margin parameter, number of pixels on each side to leave free */
static void
computeLayout(int cs[], const int avail, const int minsize, const int prefsize, const int align, const int margin)
{
    if (avail >= prefsize + 2*margin) {
        /* We have more room than required to give this item its preferred size. */
        cs[0] = margin + (avail - 2*margin - prefsize)*align/2;
        cs[1] = prefsize;
    } else if (avail >= minsize + 2*margin) {
        /* We do not have enough room to give it its preferred size, but we have
           more than its minimum. Thus, expand it to full size. */
        cs[0] = margin;
        cs[1] = avail - margin;
    } else if (avail >= minsize) {
        /* We have enough room to give it its minimum size when we reduce the margin */
        cs[0] = (avail - minsize) / 2;
        cs[1] = minsize;
    } else {
        /* We're even smaller than its minimum size. Give it everything we have. */
        cs[0] = 0;
        cs[1] = avail;
    }
}

/** Constructor.
    \param content Contained widget
    \param align   Alignment parameter. X is 0=left, 1=center, 2=right, Y is 0=top, 1=center, 2=bottom
    \param margin  Margin. Number of pixels to leave free on both sides */
UIAlignedWidget::UIAlignedWidget(UIBaseWidget& content, GfxPoint align, GfxPoint margin)
    : align(align),
      margin(margin)
{
    addChildWidget(content, 0);
}

UIAlignedWidget::~UIAlignedWidget()
{ }

void
UIAlignedWidget::onResize(GfxRect /*oldsize*/, bool by_layout)
{
    if (UIBaseWidget* fc = getFirstChild()) {
        LayoutInfo info;
        fc->getLayoutInfo(info);

        int xs[2], ys[2];
        computeLayout(xs, extent.w, info.min_size.x, info.pref_size.x, align.x, margin.x);
        computeLayout(ys, extent.h, info.min_size.y, info.pref_size.y, align.y, margin.y);
        fc->setExtent(GfxRect(extent.x + xs[0], extent.y + ys[0], xs[1], ys[1]), by_layout);
    }
}

void
UIAlignedWidget::getLayoutInfo(LayoutInfo& info)
{
    if (UIBaseWidget* fc = getFirstChild()) {
        fc->getLayoutInfo(info);
        info.pref_size.x += 2*margin.x;
        info.pref_size.y += 2*margin.y;
        info.min_size.x += 2*margin.x;
        info.min_size.y += 2*margin.y;
    } else {
        info.pref_size = info.min_size = GfxPoint(0, 0);
        info.flags     = li_NoLayout;
    }
}

void
UIAlignedWidget::drawContent(GfxCanvas& /*canvas*/)
{ }
