/**
  *  \file util/conf.cc
  *  \brief Configuration Data Management
  *
  *  \todo Exception handling must be correctly specified and implemented.
  *  Right now we throw range_error's or ignore errors.
  *
  *  \todo Associate some "source" parameter instead of just a
  *  "was_set" flag with each option, to implement things like "this
  *  option was set in the system configuration file" or "save only
  *  the options modified on the config screen to the game config
  *  file".
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "util/conf.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/misc.h"
#include "cpluslib/format.h"

/********************************* Config ********************************/

/** Constructor. Make blank object. */
Config::Config()
{ }

/** Destructor. */
Config::~Config()
{ }

/** Add option.
    \param option The option. Its lifetime must be the same or longer as the
                  Config object's. */
void
Config::addConfigOption(ConfigOption& option)
{
    options.push_back(&option);
}

/** Get option by name.
    \param name Name of config option
    \return the option, or null if none matches */
ConfigOption*
Config::getConfigOption(const string_t name)
{
    for (std::vector<ConfigOption*>::iterator i = options.begin(); i != options.end(); ++i)
        if (strCaseCmp(name, (*i)->getName()) == 0)
            return *i;
    return 0;
}

/** Mark all options as unset. */
void
Config::markAllOptionsUnset()
{
    for (std::vector<ConfigOption*>::iterator i = options.begin(); i != options.end(); ++i)
        (*i)->markUnset();
}

/****************************** ConfigOption *****************************/

/** Constructor. Constructs an object and adds it to the container.
    \param container Config object
    \param name Name of option. Lifetime must be the same or longer as
                this ConfigOption object's, i.e. static. */
ConfigOption::ConfigOption(Config& container, const char* name)
    : name(name), was_set(false)
{
    container.addConfigOption(*this);
}

/** Get name of this option. */
const char*
ConfigOption::getName() const
{
    return name;
}

/****************************** ValueParser ******************************/

/** Parse comma-separated list into array.
    \param value [in] String value
    \param array [out] The array
    \param dim   [in] Number of elements in array */
void
ValueParser::parseArray(string_t value, int32 array[], int dim)
{
    int count = 0;
    int32 last_value;
    string_t::size_type n;
    while (count < dim-1 && (n = value.find(',')) != string_t::npos) {
        last_value = parse(strTrim(value.substr(0, n)));
        value.erase(0, n+1);
        array[count++] = last_value;
    }
    last_value = parse(strTrim(value));
    array[count++] = last_value;

    while (count < dim)
        array[count++] = last_value;
}

string_t
ValueParser::toStringArray(const int32 array[], int dim)
{
    string_t result = toString(array[0]);
    for (int i = 1; i < dim; ++i) {
        result += ",";
        result += toString(array[i]);
    }
    return result;
}

/***************************** ValueIntParser ****************************/

ValueIntParser ValueIntParser::instance;

/** Constructor. */
ValueIntParser::ValueIntParser()
{ }

/** Parse integer. */
int32
ValueIntParser::parse(string_t value)
{
    int32 result;
    if (!stringValue(value, result))
        throw std::range_error(_("Invalid number"));
    return result;
}

string_t
ValueIntParser::toString(int32 value)
{
    return format("%d", value);
}

/**************************** ValueBoolParser ****************************/

ValueBoolParser ValueBoolParser::instance;

ValueBoolParser::ValueBoolParser()
{ }

int32
ValueBoolParser::parse(string_t value)
{
    if (stringMatch("No", value) || stringMatch("False", value))
        return 0;
    if (stringMatch("Yes", value) || stringMatch("True", value))
        return 1;
    if (stringMatch("Allies", value))
        return 2;
    if (stringMatch("External", value))
        return 3;

    int32 tmp;
    if (stringValue(value, tmp))
        return tmp;

    return 1; /* arbitrary */
}

string_t
ValueBoolParser::toString(int32 value)
{
    switch (value) {
     case 0:
        return "No";
     case 1:
     default: /* arbitrary */
        return "Yes";
     case 2:
        return "Allies";
     case 3:
        return "External";
    }
}

/*************************** ValueBitsetParser ***************************/

static bool
nextWord(const char*& tpl, string_t& word)
{
    using namespace std;
    if (*tpl == 0) {
        // end reached
        return false;
    } else if (const char* p = strchr(tpl, ',')) {
        // intermediate word
        word.assign(tpl, p - tpl);
        tpl = p+1;
        return true;
    } else {
        // final word
        word.assign(tpl);
        tpl += strlen(tpl);
        return true;
    }
}

/** Constructor.
    \param tpl Template naming all the bits, in the form "zero,one,two,four,eight". */
ValueBitsetParser::ValueBitsetParser(const char* tpl)
    : tpl(tpl)
{ }

int32
ValueBitsetParser::parse(string_t value)
{
    int32 result = 0;
    do {
        string_t item = strTrim(strFirst(value, ","));
        int32 tmp;
        if (item.size() == 0) {
            // blank field
        } else if (stringValue(item, tmp)) {
            // numeric field
            result |= tmp;
        } else {
            // word field
            const char* p = tpl;
            string_t ele;
            int32 bit = 1;
            while (nextWord(p, ele)) {
                if (strCaseCmp(ele, item) == 0) {
                    result |= bit;
                    break;
                }
                bit <<= 1;
            }
        }
    } while (strRemove(value, ","));
    return result;
}

string_t
ValueBitsetParser::toString(int32 value)
{
    string_t result;
    const char* p = tpl;
    string_t ele;
    int32 bit = 1;
    while (nextWord(p, ele)) {
        if (value & bit) {
            if (result.size())
                result.append(",");
            result.append(ele);
        }
        bit <<= 1;
    }
    return result;
}

/*************************** ConfigStringOption **************************/

ConfigStringOption::ConfigStringOption(Config& container, const char* name)
    : ConfigOption(container, name)
{ }

ConfigStringOption::~ConfigStringOption()
{ }

void
ConfigStringOption::set(string_t value)
{
    this->value = value;
}

string_t
ConfigStringOption::toString() const
{
    return value;
}

string_t
ConfigStringOption::operator()() const
{
    return value;
}

/****************************** ConfigAlias ******************************/


ConfigAlias::ConfigAlias(Config& container, const char* name, ConfigOption& other)
    : ConfigOption(container, name),
      other(other)
{ }

void
ConfigAlias::set(string_t value)
{
    other.set(value);
}

string_t
ConfigAlias::toString() const
{
    return other.toString();
}
