/**
  *  \file util/xmlread.cc
  *  \brief Simple XML Reader
  *
  *  Implements the XmlReader class.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "util/xmlread.h"
#include "cpluslib/utf8.h"
#include "cpluslib/assert.h"

/** Constructor. Note that this will already read the first bytes from
    the stream to detect the character encoding.
    \param s Stream to read from */
XmlReader::XmlReader(Stream& s)
    : s(s),
      buf_ptr(0),
      buf_fill(0),
      eof(false),
      encoding(Utf8),
      codepage(),
      state(Main),
      current_char(0),
      tag(),
      name(),
      value()
{
    refillBuffer();
    detectEncoding();
    readNextChar();
}

XmlReader::~XmlReader()
{ }

/** Read a token. */
XmlReader::Token
XmlReader::read()
{
    while (1) {
        Token t = readNextToken();
        if (t == Null) {
            // ignore
        } else if (t == Text && value.empty()) {
            // ignore
        } else {
            // accept
            return t;
        }
    }
}

/** Get current tag name. For TagStart, TagAttribute, TagEnd, this is
    the tag name. For PIStart, PIAttribute, this is the target name. */
string_t
XmlReader::getTag() const
{
    return tag;
}

/** Get current attribute name. Valid for TagAttribute and PIAttribute. */
string_t
XmlReader::getName() const
{
    return name;
}

/** Get current attribute value or text. Valid for TagAttribute and
    PIAttribute, and for Text. */
string_t
XmlReader::getValue() const
{
    return value;
}

/** Expand an entity reference. Can be overridden to support
    additional names. This implements numeric references (decimal,
    hex) and the standard XML entity names (lt, gt, apos, quot, amp).
    \param e entity name
    \return resulting string */
string_t
XmlReader::expandEntity(string_t e)
{
    if (e.size() == 0) {
        return string_t();
    } else if (e[0] == '#') {
        string_t result;
        if (e.size() > 2 && e[1] == 'x') {
            appendUtf8(result, std::strtoul(e.c_str()+2, 0, 16));
        } else {
            appendUtf8(result, std::strtoul(e.c_str()+1, 0, 10));
        }
        return result;
    } else if (e == "lt") {
        return "<";
    } else if (e == "gt") {
        return ">";
    } else if (e == "apos") {
        return "\'";
    } else if (e == "quot") {
        return "\"";
    } else if (e == "amp") {
        return "&";
    } else {
        return e;
    }
}


/** Read a (pseudo)token. */
XmlReader::Token
XmlReader::readNextToken()
{
    // End of file reached?
    if (current_char < 0)
        return Eof;

    // Regular processing
    switch (state) {
     case Main:
        if (current_char == '<') {
            // Tag
            readNextChar();
            readCharacterSequence(WhitespaceChar, 0);
            if (current_char == '?') {
                // "<?NAME ...."
                readNextChar();
                tag.clear();
                readCharacterSequence(WhitespaceChar, 0);
                readCharacterSequence(~(WhitespaceChar | GTChar | QuesChar | GTChar), &tag);
                state = InPI;
                return PIStart;
            } else if (current_char == '/') {
                // "</NAME>"
                readNextChar();
                tag.clear();
                readCharacterSequence(WhitespaceChar, 0);
                readCharacterSequence(~(WhitespaceChar | GTChar), &tag);    // read tag name
                readCharacterSequence(~GTChar, 0);                          // ignore up to closing angle bracket
                readNextChar();                                             // read closing angle bracket
                return TagEnd;
            } else if (current_char == '!') {
                // "<!-- .... -->" or "<![CDATA[....]]>"
                readNextChar();
                if (current_char == '[') {
                    string_t tmp;
                    readNextChar();
                    readCharacterSequence(~(LBracketChar | ProcessQuotes), &tmp);
                    if (tmp == "CDATA") {
                        // "<![CDATA[....]]>"
                        readNextChar();                                     // skip the final "["
                        value.clear();
                        readUnparsedCharacterData(value);
                        return Text;
                    } else {
                        // Any other directive (INCLUDE and IGNORE can appear in DTDs)
                        readCharacterSequence(~GTChar, 0);
                        readNextChar();
                        return Error;
                    }
                } else if (current_char == '-') {
                    // possible comment
                    if (readCharacterSequence(MinusChar, 0) >= 2) {
                        // "<!--". Comment ends at "-->". XML-1.1 forbids "--" within a comment,
                        // so we treat that as the comment end (although we allow more than two
                        // "--"s, which is also not permitted).
                        do {
                            readCharacterSequence(~(MinusChar | ProcessQuotes), 0);
                        } while (current_char == '-' && readCharacterSequence(MinusChar, 0) < 2);
                    }
                    readCharacterSequence(~GTChar, 0);
                    readNextChar();
                    return Comment;
                } else {
                    // anything else, e.g. "<!DOCTYPE>" block: skip until '>'
                    readCharacterSequence(~GTChar, 0);
                    readNextChar();
                    return Null;
                }
            } else {
                // "<NAME....>"
                tag.clear();
                readCharacterSequence(~(WhitespaceChar | GTChar), &tag);    // read tag name
                state = InTag;
                return TagStart;
            }
        } else {
            // Text
            value.clear();
            readCharacterSequence(~(LTChar | ProcessQuotes), &value);
            if (value.find_first_not_of("\r\n\t ") == value.npos)
                value.clear();
            else
                value = expandEntities(value);
            return Text;
        }
        break;

     case InTag:
        readCharacterSequence(WhitespaceChar, 0);
        if (current_char == '/') {
            // Tag ends and is immediately closed
            readCharacterSequence(~GTChar, 0);
            readNextChar();
            state = Main;
            return TagEnd;
        } else if (current_char == '>') {
            // Tag ends normally
            readNextChar();
            state = Main;
            return Null;
        } else {
            // Tag attribute
            name.clear();
            readCharacterSequence(~(WhitespaceChar | EQChar | GTChar | SlashChar), &name);
            readCharacterSequence(WhitespaceChar, 0);
            if (current_char == '=') {
                readNextChar();
                value.clear();
                readCharacterSequence(WhitespaceChar, 0);
                readCharacterSequence(~(WhitespaceChar | GTChar| SlashChar), &value);
                value = expandEntities(value);
            } else {
                value = name;
            }
            return TagAttribute;
        }

     case InPI:
        readCharacterSequence(WhitespaceChar, 0);
        if (current_char == '?' || current_char == '>') {
            // PI ends
            readCharacterSequence(~GTChar, 0);
            readNextChar();
            state = Main;
            return Null;
        } else {
            // PI attribute
            name.clear();
            readCharacterSequence(~(WhitespaceChar | EQChar | QuesChar | GTChar), &name);
            readCharacterSequence(WhitespaceChar, 0);
            if (current_char == '=') {
                readNextChar();
                value.clear();
                readCharacterSequence(WhitespaceChar, 0);
                readCharacterSequence(~(WhitespaceChar | QuesChar | GTChar), &value);
            } else {
                value = name;
            }

            // Check for encoding attribute. Only when in UTF-8 or codepage mode (that is, we are not
            // already interpreting this as a 16-bits-per-character file, and have not
            // yet seen another encoding name).
            if ((encoding == Utf8 || encoding == Codepage) && tag == "xml" && name == "encoding") {
                if (CharacterSet::isCharsetNameUtf8(value)) {
                    // UTF-8.
                    encoding = Utf8;
                } else {
                    // Codepage. Unsupported codepages are mapped to default (=latin1).
                    encoding = Codepage;
                    codepage = CharacterSet::getCharsetByName(value);
                }
            }

            return PIAttribute;
        }
    }

    return Error;
}

/** Read next character. Reads a Unicode character into current_char. 
    Handles all encodings. */
void
XmlReader::readNextChar()
{
    int a, b, c;
    switch (encoding) {
     case Codepage:
        a = readByte();
        if (a >= 0) {
            // Valid character
            current_char = codepage.getUnicode(a);
        } else {
            // End of file
            current_char = a;
        }
        break;

     case Utf8:
        a = readByte();
        if (a >= 0xC0 && a < 0xE0) {
            // two-byte rune
            b = readByte();
            current_char = ((a & 0x1F) << 6) + (b & 0x3F);
        } else if (a >= 0xE0 && a < 0xF0) {
            // three-byte rune
            b = readByte();
            c = readByte();
            current_char = ((a & 0x0F) << 12) + ((b & 0x3F) << 6) + (c & 0x3F);
        } else {
            // return as-is. This includes ASCII, invalid characters and EOF
            current_char = a;
        }
        break;

     case Utf16BE:
        a = readByte();
        b = readByte();
        current_char = b >= 0 ? 256*a + b : -1;
        break;

     case Utf16LE:
        a = readByte();
        b = readByte();
        current_char = b >= 0 ? 256*b + a : -1;
        break;
    }
}

/** Read byte.
    \return the byte [0,255], -1 on error */
int
XmlReader::readByte()
{
    if (buf_ptr >= buf_fill)
        refillBuffer();
    return eof ? -1 : uint8_t(buffer[buf_ptr++]);
}

/** Fill read buffer. Sets buf_ptr/buf_fill/eof.
    \pre buffer completely read */
void
XmlReader::refillBuffer()
{
    ASSERT(buf_ptr >= buf_fill);
    if (!eof) {
        buf_ptr = 0;
        buf_fill = s.read(buffer, sizeof(buffer));
        if (buf_fill == 0)
            eof = true;
    }
}

/** Skip bytes without reading.
    \param n Number of bytes to skip
    \pre buffer contains at least n unread bytes */
inline void
XmlReader::skip(std::size_t n)
{
    ASSERT(buf_fill - buf_ptr >= n);
    buf_ptr += n;
}

/** Read-ahead byte without removing from buffer
    \param n Index of byte to check (0=the one to be read next, 1=after that, ...)
    \pre buffer contains at least n+1 unread bytes */
inline uint8_t
XmlReader::peek(std::size_t n)
{
    ASSERT(buf_fill - buf_ptr >= n);
    return buffer[buf_ptr + n];
}

/** Detect encoding. Assumes pre-filled buffer, otherwise it may not
    detect everything possible. */
void
XmlReader::detectEncoding()
{
    /* Encoding detection patterns according to W3C XML-1.1 document:
         00 00 FE FF   UCS-4, big-endian machine (1234 order)
         FF FE 00 00   UCS-4, little-endian machine (4321 order)
         00 00 FF FE   UCS-4, unusual octet order (2143)
         FE FF 00 00   UCS-4, unusual octet order (3412)
         FE FF         UTF-16, big-endian
         FF FE         UTF-16, little-endian
         EF BB BF      UTF-8

         00 00 00 3C   UCS-4 BE (1234)
         3C 00 00 00   UCS-4 LE (4321)
         00 00 3C 00   UCS-4 unusual (2143)
         00 3C 00 00   UCS-4 unusual (3412)
         00 3C 00 3F   UTF-16 BE
         3C 00 3F 00   UTF-16 LE
         3C 3F 78 6D   UTF-8, ISO 646, ASCII, etc.

       We want to support UTF-16 and UTF-8, so we remain at
         hh ll         UTF-16 BE
         ll hh         UTF-16 LE
         EF BB BF      UTF-8
       with hhll being a valid Unicode character, in particular 0xFEFF or,
       for simplicity, any ASCII character (including 0x3C, but also newline
       and spaces). */
    std::size_t avail = buf_fill - buf_ptr;
    if (avail >= 3 && peek(0) == 0xEF && peek(1) == 0xBB && peek(2) == 0xBF) {
        // UTF-8 BOM
        encoding = Utf8;
        skip(3);
    } else if (avail >= 2 && peek(0) == 0xFE && peek(1) == 0xFF) {
        // UTF-16BE BOM
        encoding = Utf16BE;
        skip(2);
    } else if (avail >= 2 && peek(0) == 0xFF && peek(1) == 0xFE) {
        // UTF-16LE BOM
        encoding = Utf16LE;
        skip(2);
    } else if (avail >= 2 && peek(0) == 0 && peek(1) != 0) {
        // UTF-16BE-encoded ASCII character
        encoding = Utf16BE;
    } else if (avail >= 2 && peek(0) != 0 && peek(1) == 0) {
        // UTF-16LE-encoded ASCII character
        encoding = Utf16LE;
    } else {
        // anything else
        encoding = Utf8;
    }
}

/** Read character sequence.
    \param cl     [in] Character classes to accept.
    \param accum  [out,optional] Characters will be stored here
    \return Number of characters read */
int32_t
XmlReader::readCharacterSequence(int cl, string_t* accum)
{
    int32_t count = 0;
    int32_t current_quote = 0;
    while (current_char >= 0) {
        // Figure out this character's class
        int my_class;
        if (current_char <= ' ' || current_char == 0x85 /* NEXT LINE */ || current_char == 0x2028 /* LINE SEPARATOR */)
            my_class = WhitespaceChar;
        else if (current_char == '<')
            my_class = LTChar;
        else if (current_char == '>')
            my_class = GTChar;
        else if (current_char == '=')
            my_class = EQChar;
        else if (current_char == '!')
            my_class = BangChar;
        else if (current_char == '?')
            my_class = QuesChar;
        else if (current_char == '/')
            my_class = SlashChar;
        else if (current_char == '-')
            my_class = MinusChar;
        else if (current_char == '[')
            my_class = LBracketChar;
        else
            my_class = AnyChar;

        // Quoting
        if (current_quote != 0 && current_char == current_quote) {
            // Quoting ends
            current_quote = 0;
        } else if ((cl & ProcessQuotes) != 0 && (current_char == '\"' || current_char == '\'')) {
            // Quoting starts
            current_quote = current_char;
        } else if ((cl & my_class) != 0 || current_quote != 0) {
            // Matching class or quoted
            if (accum)
                appendUtf8(*accum, current_char);
            ++count;
        } else {
            // Non-matching class
            break;
        }

        // Consume this character
        readNextChar();
    }
    return count;
}

/** Read content of a "<![CDATA" block. Assumes everything up to and
    including the opening bracket has already been read. Reads
    everything up to and including the final "]]>". */
void
XmlReader::readUnparsedCharacterData(string_t& out)
{
    // Parse content
    int numRBrack = 0;
    while (current_char >= 0) {
        if (numRBrack == 0) {
            if (current_char == ']') {
                // First terminator
                numRBrack = 1;
            } else {
                appendUtf8(out, current_char);
            }
        } else if (numRBrack == 1) {
            if (current_char == ']') {
                // Second terminator
                numRBrack = 2;
            } else {
                // Single "]" does not terminate: append it, and append next character
                out += ']';
                appendUtf8(out, current_char);
            }
        } else if (numRBrack == 2) {
            if (current_char == '>') {
                // Third terminator
                readNextChar();
                break;
            } else if (current_char == ']') {
                // Third "]" means at least one of them belongs to the text
                out += ']';
            } else {
                // Third terminator missing
                out += "]]";
                appendUtf8(out, current_char);
            }
        }
        readNextChar();
    }
}

/** Expand entity-references within a string. Users can override
    expandEntity to define the entity names to accept.
    \param s the string
    \return Updated string */
string_t
XmlReader::expandEntities(const string_t s)
{
    string_t out;
    string_t::size_type pos = 0;
    string_t::size_type n, q;
    while ((n = s.find('&', pos)) != string_t::npos) {
        // Locate end of entity reference
        q = s.find(';', n);
        if (q == string_t::npos)
            break;

        out.append(s, pos, n-pos);
        out.append(expandEntity(string_t(s, n+1, q-n-1)));
        pos = q+1;
    }
    out.append(s.substr(pos));
    return out;
}
